"""Base classes and assorted functions for beep structuring datapaths.
"""

import os
import abc
import json
import copy
import time
import itertools

import pandas as pd
import numpy as np
from scipy import integrate
from monty.json import MSONable
from monty.io import zopen
from monty.serialization import dumpfn

from beep import tqdm
from beep import MODULE_DIR
from beep.conversion_schemas import (
    STRUCTURE_DTYPES,
)

from beep.utils import parameters_lookup
from beep import logger, VALIDATION_SCHEMA_DIR, PROTOCOL_PARAMETERS_DIR
from beep.structure.validate import SimpleValidator


class BEEPDatapath(abc.ABC, MSONable):
    """The base class for all beep datapaths.

    BEEPDatapaths handle all BEEP structuring data, including

    - holding raw data
    - validating data
    - normal and diagnostic cycle interpolation
    - diagnostic extraction
    - automatic structuring
    - normal and diagnostic cycle summarization
    - determining cycle-capacity and capacity-cycle relationships
    - determining cycle lifetimes
    - saving and loading processed (structured) runs* from static json files
        - *including legacy json serialized with earlier versions of BEEP

    Note that BEEPDatapath does NOT handle featurization or ML modelling, only data organization and munging.


    BEEPDatapath is an abstract base class requiring a child class to implement the following methods:
    - from_file: Take raw cycler output files and convert them to a BEEPDatapath object. It should
        return a BEEPDatapath object with the correct data types (specified in __init__).


    Attributes:

        Important/Very common attributes:
            Each BEEPDatapath will have a maximum of 5 *very important* structured attributes dataframes:
            - structured_summary (pd.DataFrame): A summary of the cycles
            - structured_data (pd.DataFrame): The interpolated cycles
            - diagnostic_data (pd.DataFrame): The interpolated diagnostic cycles
            - diagnostic_summary (pd.DataFrame): A summary of diagnostic cycles
            - raw_data (pd.DataFrame): All raw data from the raw input file.

        Less important attributes:
            - structuring_parameters: A dictionary of the structuring parameters used to structure
                this object.
            - metadata (BEEPDatapath.CyclerRunMetadata): An object holding all metadata.
            - paths (dict): A mapping of {descriptor: full_path or [paths]} for all files related to this datapath.
                This can include things like "raw", "metadata", "structured", as well as other paths (e.g., "eis").
            - schema: Validation schema used to validate the raw ingested data.

        Private:
            - _is_legacy (bool): Whether this file is loaded from a legacy ProcessedCyclerRun. Some operations are
                not supported for legacy-loaded processed/structured files.
            - _aggregation (dict): Specifies the pandas aggregation columns order for normal cycler data
            - _diag_aggregation (dict): Specifies the pandas aggregation columns order for diagnostic data
            - _summary_cols (list): The ordering of summary columns for normal cycler data.
            - _diag_summary_cols (list): The ordering of summary columns for diagnostic data.

    """

    IMPUTABLE_COLUMNS = ["temperature", "internal_resistance"]

    class StructuringDecorators:
        """
        Internal container class for decorators related to data structuring.
        """

        @classmethod
        def must_be_structured(cls, func):
            """
            Decorator to check if the datapath has been structured.

            Args:
                func: A function or method.
            Returns:
                A wrapper function for the input function/method.
            """

            def wrapper(*args, **kwargs):
                if not args[0].is_structured:
                    raise RuntimeError(
                        f"{args[0].__class__.__name__} has not been structured! Run .structure(*args)."
                    )
                else:
                    return func(*args, **kwargs)

            return wrapper

        @classmethod
        def must_not_be_legacy(cls, func):
            """
            Decorator to check if datapath has been serialized from legacy,
            as some operations depend on data which is not in legacy serialized files.

            Args:
                func: A function or method

            Returns:
                A wrapper function for the input function/method.

            """

            def wrapper(*args, **kwargs):
                if args[0]._is_legacy:
                    raise ValueError(
                        f"{args[0].__class__.__name__} is deserialized from a legacy file! Operation not allowed."
                    )
                else:
                    return func(*args, **kwargs)

            return wrapper

    class CyclerRunMetadata:
        """
        Container class for cycler metadata. Does not change based on whether the instance is structured
        or unstructured.

        Attributes:
            raw (dict): The raw metadata dict, containing all input metadata.
            barcode (str): The barcode
            protocol (any, json serializable): The protocol used by the cycler for this run
            channel_id (int): Channel id number
        """

        def __init__(self, metadata_dict):
            """

            Args:
                metadata_dict (dict): A dictionary which can contain arbitrary metadata.
            """

            # Keep all the common metadata attr-accessible
            self.barcode = metadata_dict.get("barcode")
            self.protocol = metadata_dict.get("protocol")
            self.channel_id = metadata_dict.get("channel_id")

            # Extra metadata will always be in .raw
            self.raw = metadata_dict

    def __init__(self, raw_data, metadata, paths=None, schema=None, impute_missing=True):
        """

        Args:
            raw_data (pd.DataFrame): A pandas dataframe of raw data. Must contain the columns specified in
                summary_cols.
            metadata (dict): A dictionary for metadata. Should contain keys specified in CyclerRunMetadata, but does not
                have to.
            paths ({str: str, Pathlike}): Should contain "raw" and "metadata" keys, even if they are the same
                filepath.
            schema (str): the name of the validation schema file to use. Should be located in the
                VALIDATION_SCHEMA_DIR directory, or can alternatively be an absolute filepath.
            impute_missing (bool): Impute missing columns such as temperature and internal_resistance if they
                are not included in raw_data.
        """
        self.raw_data = raw_data

        # paths may include "raw", "metadata", and "structured", as well as others.
        if paths:
            for path_ref, path in paths.items():
                if path and not os.path.isabs(path):
                    raise ValueError(f"{path_ref}: '{path}' is not absolute! All paths must be absolute.")
            self.paths = paths
        else:
            self.paths = {"raw": None}

        if impute_missing:
            # support legacy operation
            if raw_data is not None:
                for col in self.IMPUTABLE_COLUMNS:
                    if col not in raw_data:
                        raw_data[col] = np.NaN

        if not schema:
            self.schema = schema
        else:
            if os.path.isabs(schema):
                abs_schema = schema
            else:
                abs_schema = os.path.join(VALIDATION_SCHEMA_DIR, schema)
            # TODO this dependence on a file path should be removed in the case of reloading an existing file
            # one solution would be to move this logic into the cycler subclasses and then pass the schema as a dict
            if os.path.exists(abs_schema):
                self.schema = abs_schema
            elif os.path.exists(os.path.join(VALIDATION_SCHEMA_DIR, os.path.split(schema)[-1])):
                self.schema = os.path.join(VALIDATION_SCHEMA_DIR, os.path.split(schema)[-1])
            else:
                raise FileNotFoundError(f"The schema file specified for validation could not be found: {schema}.")

        self.structured_summary = None     # equivalent of PCR.summary
        self.structured_data = None        # equivalent of PCR.cycles_interpolated
        self.diagnostic_data = None        # equivalent of PCR.diagnostic_interpolated
        self.diagnostic_summary = None     # same name as in PCR

        self.metadata = self.CyclerRunMetadata(metadata)

        self._is_legacy = False

        # Setting aggregation/column ordering based on whether columns are present
        self._aggregation = {
            "cycle_index": "first",
            "discharge_capacity": "max",
            "charge_capacity": "max",
            "discharge_energy": "max",
            "charge_energy": "max",
            "internal_resistance": "last",
        }

        if self.raw_data is not None and "temperature" in self.raw_data.columns:
            self._aggregation["temperature"] = ["max", "mean", "min"]

            self._summary_cols = [
                "cycle_index",
                "discharge_capacity",
                "charge_capacity",
                "discharge_energy",
                "charge_energy",
                "dc_internal_resistance",
                "temperature_maximum",
                "temperature_average",
                "temperature_minimum",
                "date_time_iso",
            ]
        else:
            self._summary_cols = [
                "cycle_index",
                "discharge_capacity",
                "charge_capacity",
                "discharge_energy",
                "charge_energy",
                "dc_internal_resistance",
                "date_time_iso",
            ]

        # Ensure this step of the aggregation is placed last
        self._aggregation["date_time_iso"] = "first"

        self._diag_aggregation = copy.deepcopy(self._aggregation)
        self._diag_aggregation.pop("internal_resistance")
        self._diag_summary_cols = copy.deepcopy(self._summary_cols)
        self._diag_summary_cols.pop(5)  # internal_resistance
        self.structuring_parameters = {}

    @classmethod
    @abc.abstractmethod
    def from_file(cls, path, *args, **kwargs):
        """Go from a raw cycler output data file (or files) to a BEEPDatapath.

        Must be implemented for the BEEPDatapath child to be valid.

        Args:
            path (str, Pathlike): The path to the raw data file.

        Returns:
            (BEEPDatapath): A child class of BEEPDatapath.

        """
        raise NotImplementedError

    @StructuringDecorators.must_not_be_legacy
    def validate(self):
        """Validate the raw data.

        Returns:
            (bool) True if the raw data is valid, false otherwise.

        """

        if self.schema:
            v = SimpleValidator(schema_filename=self.schema)
        else:
            v = SimpleValidator()

        is_valid, reason = v.validate(self.raw_data)
        return is_valid, reason

    @classmethod
    def from_json_file(cls, filename):
        """Load a structured run previously saved to file.

        .json.gz files are supported.

        Loads a BEEPDatapath or (legacy) ProcessedCyclerRun structured object from json.

        Can be used in combination with files serialized with BEEPDatapath.to_json_file.

        Args:
            filename (str, Pathlike): a json file from a structured run, serialzed with to_json_file.

        Returns:
            None
        """
        with zopen(filename, "r") as f:
            d = json.load(f)

        # Add this structured file path to the paths dict
        paths = d.get("paths", {})
        paths["structured"] = os.path.abspath(filename)
        d["paths"] = paths

        return cls.from_dict(d)

    def to_json_file(self, filename, omit_raw=True):
        """Save a BEEPDatapath to disk as a json.

        .json.gz files are supported.

        Not named from_json to avoid conflict with MSONable.from_json(*)

        Args:
            filename (str, Pathlike): The filename to save the file to.
            omit_raw (bool): If True, saves only structured (NOT RAW) data.
                More efficient for saving/writing to disk.

        Returns:
            None
        """
        d = self.as_dict()
        if omit_raw:
            d.pop("raw_data")

        dumpfn(d, filename)

    @StructuringDecorators.must_not_be_legacy
    def as_dict(self):
        """Serialize a BEEPDatapath as a dictionary.

        Must not be loaded from legacy.

        Returns:
            (dict): corresponding to dictionary for serialization.

        """

        if not self.is_structured:
            summary = None
            cycles_interpolated = None
            diagnostic_summary = None
            diagnostic_interpolated = None
        else:
            summary = self.structured_summary.to_dict("list")
            cycles_interpolated = self.structured_data.to_dict("list")
            diagnostic_summary = self.diagnostic_summary.to_dict(
                "list") if self.diagnostic_summary is not None else None
            diagnostic_interpolated = self.diagnostic_data.to_dict("list") if self.diagnostic_data is not None else None

        return {
            "@module": self.__class__.__module__,
            "@class": self.__class__.__name__,

            # Vital things needed for BEEPDatapath
            "raw_data": self.raw_data.to_dict("list"),
            "metadata": self.metadata.raw,
            "paths": self.paths,

            # For backwards compatibility
            # All this data is in the "metadata" key, this is just for redundancy
            "barcode": self.metadata.barcode,
            "protocol": self.metadata.protocol,
            "channel_id": self.metadata.channel_id,

            # Structured data, expensively obtained
            "summary": summary,
            "cycles_interpolated": cycles_interpolated,
            "diagnostic_summary": diagnostic_summary,
            "diagnostic_interpolated": diagnostic_interpolated,

            # Structuring parameters (mostly for provenance)
            "structuring_parameters": self.structuring_parameters,

            # Provence for validation
            "schema_path": self.schema,

        }

    @classmethod
    def from_dict(cls, d):
        """Create a BEEPDatapath object from a dictionary.

        Args:
            d (dict): dictionary represenation.

        Returns:
            beep.structure.ProcessedCyclerRun: deserialized ProcessedCyclerRun.
        """
        for obj, dtype_dict in STRUCTURE_DTYPES.items():
            for column, dtype in dtype_dict.items():
                if d.get(obj) is not None:
                    if d[obj].get(column) is not None:
                        d[obj][column] = pd.Series(d[obj][column], dtype=dtype)

        paths = d.get("paths", None)
        schema = d.get("schema_path", None)

        # support legacy operations
        # support loads when raw_data not available
        if any([k not in d for k in ("raw_data", "metadata")]):
            raw_data = None
            metadata = {k: d.get(k) for k in ("barcode", "protocol", "channel_id")}
            is_legacy = True
        else:
            raw_data = pd.DataFrame(d["raw_data"])
            metadata = d.get("metadata")
            is_legacy = False

        datapath = cls(raw_data=raw_data, metadata=metadata, paths=paths, schema=schema)
        datapath._is_legacy = is_legacy

        datapath.structured_data = pd.DataFrame(d["cycles_interpolated"])
        datapath.structured_summary = pd.DataFrame(d["summary"])

        diagnostic_summary = d.get("diagnostic_summary")
        diagnostic_data = d.get("diagnostic_interpolated")
        datapath.diagnostic_summary = diagnostic_summary if diagnostic_summary is None else pd.DataFrame(
            diagnostic_summary)
        datapath.diagnostic_data = diagnostic_data if diagnostic_data is None else pd.DataFrame(diagnostic_data)

        datapath.structuring_parameters = d.get("structuring_parameters", {})
        return datapath

    @property
    def semiunique_id(self):
        """
        An id that can identify the state of this datapath without complications
        associated with hashing dataframes.

        Returns:
            (str): A semiunique id for this datapath.

        """
        s = f"barcode:{self.metadata.barcode}-" \
            f"channel:{self.metadata.channel_id}-" \
            f"protocol:{self.metadata.protocol}-" \
            f"schema:{self.schema}-" \
            f"structured:{self.is_structured}-" \
            f"legacy:{self._is_legacy}"

        raw = self.paths.get("raw", None)
        structured = self.paths.get("structured", None)
        s += f"-raw_path:{raw}-structured_path:{structured}"
        return s

    @StructuringDecorators.must_not_be_legacy
    def structure(self,
                  v_range=None,
                  resolution=1000,
                  diagnostic_resolution=500,
                  nominal_capacity=1.1,
                  full_fast_charge=0.8,
                  diagnostic_available=False,
                  charge_axis='charge_capacity',
                  discharge_axis='voltage',
                  ):
        """

        Args:
            v_range ([int, int]): range of voltages for cycle interpolation.
            resolution (int): resolution for cycle interpolation.
            diagnostic_resolution (int): number of datapoints per step for
                interpolating diagnostic cycles.
            nominal_capacity (float): nominal capacity for summary stats.
            full_fast_charge (float): full fast charge for summary stats.
            diagnostic_available (dict): project metadata for processing
                diagnostic cycles correctly.
            charge_axis (str): Column to interpolate charge step
            discharge_axis (str): Column to interpolate discharge step
        """
        logger.info(f"Beginning structuring along charge axis '{charge_axis}' and discharge axis '{discharge_axis}'.")

        if diagnostic_available:
            self.diagnostic_summary = self.summarize_diagnostic(
                diagnostic_available
            )
            self.diagnostic_data = self.interpolate_diagnostic_cycles(
                diagnostic_available, diagnostic_resolution
            )

        self.structured_data = self.interpolate_cycles(
            v_range=v_range,
            resolution=resolution,
            diagnostic_available=diagnostic_available,
            charge_axis=charge_axis,
            discharge_axis=discharge_axis
        )

        self.structured_summary = self.summarize_cycles(
            nominal_capacity=nominal_capacity,
            full_fast_charge=full_fast_charge,
            diagnostic_available=diagnostic_available
        )

        self.structuring_parameters = {
            "v_range": v_range,
            "resolution": resolution,
            "diagnostic_resolution": diagnostic_resolution,
            "nominal_capacity": nominal_capacity,
            "full_fast_charge": full_fast_charge,
            "diagnostic_available": diagnostic_available,
            "charge_axis": charge_axis,
            "discharge_axis": discharge_axis
        }

    def autostructure(
            self,
            charge_axis='charge_capacity',
            discharge_axis='voltage',
            parameters_path=None,
    ):
        """
        Automatically run structuring based on automatically determined structuring parameters.
        The parameters are determined from the raw input file, so ensure the raw input file paths
        are in the 'paths' attribute.

        Args:
            charge_axis (str): Column to interpolate charge step
            discharge_axis (str): Column to interpolate discharge step
            parameters_path (str) Path to directory of protocol parameters files.

        Returns:
            None
        """
        parameters_path = parameters_path if parameters_path else PROTOCOL_PARAMETERS_DIR
        self.paths["protocol_parameters"] = parameters_path
        v_range, resolution, nominal_capacity, full_fast_charge, diagnostic_available = \
            self.determine_structuring_parameters(parameters_path=parameters_path)
        logger.info(f"Autostructuring determined parameters of v_range={v_range}, "
                    f"resolution={resolution}, "
                    f"nominal_capacity={nominal_capacity}, "
                    f"full_fast_charge={full_fast_charge}, "
                    f"diagnostic_available={diagnostic_available}")
        return self.structure(
            v_range=v_range,
            resolution=resolution,
            nominal_capacity=nominal_capacity,
            full_fast_charge=full_fast_charge,
            diagnostic_available=diagnostic_available,
            charge_axis=charge_axis,
            discharge_axis=discharge_axis
        )

    @StructuringDecorators.must_not_be_legacy
    def unstructure(self):
        """
        Cleanly remove all structuring data, to restructure using different parameters.

        Returns:
            None
        """
        self.structured_data = None
        self.diagnostic_data = None
        self.structured_summary = None
        self.diagnostic_summary = None
        self.structuring_parameters = {}
        logger.debug("Structure has been reset.")

    def interpolate_step(
            self,
            v_range,
            resolution,
            step_type="discharge",
            reg_cycles=None,
            axis="voltage",
            desc=None
    ):
        """
        Gets interpolated cycles for the step specified, charge or discharge.

        Args:
            v_range ([Float, Float]): list of two floats that define
                the voltage interpolation range endpoints.
            resolution (int): resolution of interpolated data.
            step_type (str): which step to interpolate i.e. 'charge' or 'discharge'
            reg_cycles (list): list containing cycle indicies of regular cycles
            axis (str): which column to use for interpolation
            desc (str): Description to print to tqdm column.

        Returns:
            pandas.DataFrame: DataFrame corresponding to interpolated values.
        """

        if not desc:
            desc = \
                f"Interpolating {step_type} ({v_range[0]} - {v_range[1]})V " \
                f"({resolution} points)"

        if step_type == "discharge":
            step_filter = step_is_dchg
        elif step_type == "charge":
            step_filter = step_is_chg
        else:
            raise ValueError("{} is not a recognized step type")
        incl_columns = [
            "test_time",
            "voltage",
            "current",
            "charge_capacity",
            "discharge_capacity",
            "charge_energy",
            "discharge_energy",
            "internal_resistance",
            "temperature",
        ]
        all_dfs = []
        cycle_indices = self.raw_data.cycle_index.unique()
        cycle_indices = sorted([c for c in cycle_indices if c in reg_cycles])

        for cycle_index in tqdm(cycle_indices, desc=desc):
            # Use a cycle_index mask instead of a global groupby to save memory
            new_df = (
                self.raw_data.loc[self.raw_data["cycle_index"] == cycle_index]
                    .groupby("step_index")
                    .filter(step_filter)
            )
            if new_df.size == 0:
                continue

            if axis in ["charge_capacity", "discharge_capacity"]:
                axis_range = [self.raw_data[axis].min(),
                              self.raw_data[axis].max()]
                new_df = interpolate_df(
                    new_df,
                    axis,
                    field_range=axis_range,
                    columns=incl_columns,
                    resolution=resolution,
                )
            elif axis == "test_time":
                axis_range = [new_df[axis].min(), new_df[axis].max()]
                new_df = interpolate_df(
                    new_df,
                    axis,
                    field_range=axis_range,
                    columns=incl_columns,
                    resolution=resolution,
                )
            elif axis == "voltage":
                new_df = interpolate_df(
                    new_df,
                    axis,
                    field_range=v_range,
                    columns=incl_columns,
                    resolution=resolution,
                )
            else:
                raise ValueError(f"Axis {axis} not a valid step interpolation axis.")
            new_df["cycle_index"] = cycle_index
            new_df["step_type"] = step_type
            new_df["step_type"] = new_df["step_type"].astype("category")
            all_dfs.append(new_df)

        # Ignore the index to avoid issues with overlapping voltages
        result = pd.concat(all_dfs, ignore_index=True)

        # Cycle_index gets a little weird about typing, so round it here
        result.cycle_index = result.cycle_index.round()

        return result

    def interpolate_cycles(
            self,
            v_range=None,
            resolution=1000,
            diagnostic_available=None,
            charge_axis='charge_capacity',
            discharge_axis='voltage'
    ):
        """
        Gets interpolated cycles for both charge and discharge steps.

        Args:
            v_range ([float, float]): list of two floats that define
                the voltage interpolation range endpoints.
            resolution (int): resolution of interpolated data.
            diagnostic_available (dict): dictionary containing information about
                location of diagnostic cycles
            charge_axis (str): column to use for interpolation for charge
            discharge_axis (str): column to use for interpolation for discharge

        Returns:
            (pandas.DataFrame): DataFrame corresponding to interpolated values.
        """
        if diagnostic_available:
            diag_cycles = list(
                itertools.chain.from_iterable(
                    [
                        list(range(i, i + diagnostic_available["length"]))
                        for i in diagnostic_available["diagnostic_starts_at"]
                        if i <= self.raw_data.cycle_index.max()
                    ]
                )
            )
            reg_cycles = [
                i for i in self.raw_data.cycle_index.unique() if
                i not in diag_cycles
            ]
        else:
            reg_cycles = [i for i in self.raw_data.cycle_index.unique()]

        v_range = v_range or [2.8, 3.5]

        # If any regular cycle contains a waveform step, interpolate on test_time.

        if self.raw_data[self.raw_data.cycle_index.isin(reg_cycles)]. \
                groupby(["cycle_index", "step_index"]). \
                apply(step_is_waveform_dchg).any():
            discharge_axis = 'test_time'

        if self.raw_data[self.raw_data.cycle_index.isin(reg_cycles)]. \
                groupby(["cycle_index", "step_index"]). \
                apply(step_is_waveform_chg).any():
            charge_axis = 'test_time'

        interpolated_discharge = self.interpolate_step(
            v_range,
            resolution,
            step_type="discharge",
            reg_cycles=reg_cycles,
            axis=discharge_axis,
        )
        interpolated_charge = self.interpolate_step(
            v_range,
            resolution,
            step_type="charge",
            reg_cycles=reg_cycles,
            axis=charge_axis,
        )
        result = pd.concat(
            [interpolated_discharge, interpolated_charge], ignore_index=True
        )

        return self._cast_dtypes(result, "cycles_interpolated")

    # equivalent of legacy get_summary
    def summarize_cycles(
            self,
            diagnostic_available=False,
            nominal_capacity=1.1,
            full_fast_charge=0.8,
            cycle_complete_discharge_ratio=0.97,
            cycle_complete_vmin=3.3,
            cycle_complete_vmax=3.3,
            error_threshold=1e6
    ):
        """
        Gets summary statistics for data according to cycle number. Summary data
        must be float or int type for compatibility with other methods

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            cycle_complete_discharge_ratio (float): expected ratio
                discharge/charge at the end of any complete cycle
            cycle_complete_vmin (float): expected voltage minimum achieved
                in any complete cycle
            cycle_complete_vmax (float): expected voltage maximum achieved
                in any complete cycle
            error_threshold (float): threshold to consider the summary value
                an error (applied only to specific columns that should reset
                each cycle)

        Returns:
            (pandas.DataFrame): summary statistics by cycle.

        """
        # Filter out only regular cycles for summary stats. Diagnostic summary computed separately
        if diagnostic_available:
            diag_cycles = list(
                itertools.chain.from_iterable(
                    [
                        list(range(i, i + diagnostic_available["length"]))
                        for i in diagnostic_available["diagnostic_starts_at"]
                        if i <= self.raw_data.cycle_index.max()
                    ]
                )
            )
            reg_cycles_at = [
                i for i in self.raw_data.cycle_index.unique() if
                i not in diag_cycles
            ]
        else:
            reg_cycles_at = [i for i in self.raw_data.cycle_index.unique()]

        summary = self.raw_data.groupby("cycle_index").agg(self._aggregation)

        # pd.set_option('display.max_rows', 500)
        # pd.set_option('display.max_columns', 500)
        # pd.set_option('display.width', 1000)

        summary.columns = self._summary_cols

        summary = summary[summary.index.isin(reg_cycles_at)]
        summary["energy_efficiency"] = (
                summary["discharge_energy"] / summary["charge_energy"]
        )
        summary.loc[
            ~np.isfinite(summary["energy_efficiency"]), "energy_efficiency"
        ] = np.NaN
        # This code is designed to remove erroneous energy values
        for col in ["discharge_energy", "charge_energy"]:
            summary.loc[summary[col].abs() > error_threshold, col] = np.NaN
        summary["charge_throughput"] = summary.charge_capacity.cumsum()
        summary["energy_throughput"] = summary.charge_energy.cumsum()

        # This method for computing charge start and end times implicitly
        # assumes that a cycle starts with a charge step and is then followed
        # by discharge step.
        charge_start_time = \
            self.raw_data.groupby("cycle_index", as_index=False)[
                "date_time_iso"
            ].agg("first")

        charge_finish_time = (
            self.raw_data[
                self.raw_data.charge_capacity >= nominal_capacity * full_fast_charge]
            .groupby("cycle_index", as_index=False)["date_time_iso"]
            .agg("first")
        )

        # Left merge, since some cells might not reach desired levels of
        # charge_capacity and will have NaN for charge duration
        merged = charge_start_time.merge(
            charge_finish_time, on="cycle_index", how="left"
        )

        # Charge duration stored in seconds - note that date_time_iso is only ~1sec resolution
        time_diff = np.subtract(
            pd.to_datetime(merged.date_time_iso_y, utc=True, errors="coerce"),
            pd.to_datetime(merged.date_time_iso_x, errors="coerce"),
        )
        summary["charge_duration"] = np.round(
            time_diff / np.timedelta64(1, "s"), 2)

        # Compute time since start of cycle in minutes. This comes handy
        # for featurizing time-temperature integral
        self.raw_data["time_since_cycle_start"] = pd.to_datetime(
            self.raw_data["date_time_iso"]
        ) - pd.to_datetime(
            self.raw_data.groupby("cycle_index")["date_time_iso"].transform(
                "first")
        )
        self.raw_data["time_since_cycle_start"] = (self.raw_data[
                                                       "time_since_cycle_start"] / np.timedelta64(
            1, "s")) / 60

        # Group by cycle index and integrate time-temperature
        # using a lambda function.
        if "temperature" in self.raw_data.columns:
            summary["time_temperature_integrated"] = self.raw_data.groupby(
                "cycle_index").apply(
                lambda g: integrate.trapz(g.temperature, x=g.time_since_cycle_start)
            )

        # Drop the time since cycle start column
        self.raw_data.drop(columns=["time_since_cycle_start"])

        # Determine if any of the cycles has been paused
        summary["paused"] = self.raw_data.groupby("cycle_index").apply(
            get_max_paused_over_threshold)

        # Add CV_time, CV_current, CV_capacity summary stats
        CV_time = []
        CV_current = []
        CV_capacity = []
        for cycle in summary.cycle_index:
            raw_cycle = self.raw_data.loc[self.raw_data.cycle_index == cycle]
            charge = raw_cycle.loc[raw_cycle.current > 0]
            CV = get_CV_segment_from_charge(charge)
            if CV.empty:
                logger.debug(f"Failed to extract CV segment for cycle {cycle}!")
            CV_time.append(get_CV_time(CV))
            CV_current.append(get_CV_current(CV))
            CV_capacity.append(get_CV_capacity(CV))
        summary["CV_time"] = CV_time
        summary["CV_current"] = CV_current
        summary["CV_capacity"] = CV_capacity

        summary = self._cast_dtypes(summary, "summary")

        last_voltage = self.raw_data.loc[
            self.raw_data["cycle_index"] == self.raw_data["cycle_index"].max()
            ]["voltage"]
        if (
                (last_voltage.min() < cycle_complete_vmin)
                and (last_voltage.max() > cycle_complete_vmax)
                and (
                (summary.iloc[[-1]])["discharge_capacity"].iloc[0]
                > cycle_complete_discharge_ratio
                * (summary.iloc[[-1]])["charge_capacity"].iloc[0]
                    )
        ):
            return summary
        else:
            return summary.iloc[:-1]

    # equivalent of get_interpolated_diagnostic_cycles
    def interpolate_diagnostic_cycles(
            self, diagnostic_available, resolution=1000, v_resolution=0.0005, v_delta_min=0.001
    ):
        """
        Interpolates data according to location and type of diagnostic
        cycles in the data

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
                as list, 'length' of the diagnostic in cycles and location
                of the diagnostic
            resolution (int): resolution of interpolation
            v_resolution (float): voltage delta to set for range based interpolation
            v_delta_min (float): minimum voltage delta for voltage based interpolation

        Returns:
            (pd.DataFrame) of interpolated diagnostic steps by step and cycle

        """
        # Get the project name and the parameter file for the diagnostic
        project_name_list = parameters_lookup.get_project_sequence(self.paths["raw"])
        diag_path = os.path.join(MODULE_DIR, "procedure_templates")
        v_range = parameters_lookup.get_diagnostic_parameters(
            diagnostic_available, diag_path, project_name_list[0]
        )

        # Determine the cycles and types of the diagnostic cycles
        max_cycle = self.raw_data.cycle_index.max()
        starts_at = [
            i for i in diagnostic_available["diagnostic_starts_at"] if i <= max_cycle
        ]
        diag_cycles_at = list(
            itertools.chain.from_iterable(
                [range(i, i + diagnostic_available["length"]) for i in starts_at]
            )
        )
        # Duplicate cycle type list end to end for each starting index
        diag_cycle_type = diagnostic_available["cycle_type"] * len(starts_at)
        if not len(diag_cycles_at) == len(diag_cycle_type):
            errmsg = (
                "Diagnostic cycles, {}, and diagnostic cycle types, "
                "{}, are unequal lengths".format(diag_cycles_at, diag_cycle_type)
            )
            raise ValueError(errmsg)

        diag_data = self.raw_data[self.raw_data["cycle_index"].isin(diag_cycles_at)]

        # Counter to ensure non-contiguous repeats of step_index
        # within same cycle_index are grouped separately
        diag_data.loc[:, "step_index_counter"] = 0

        for cycle_index in diag_cycles_at:
            indices = diag_data.loc[diag_data.cycle_index == cycle_index].index
            step_index_list = diag_data.step_index.loc[indices]
            diag_data.loc[indices, "step_index_counter"] = step_index_list.ne(
                step_index_list.shift()
            ).cumsum()

        group = diag_data.groupby(["cycle_index", "step_index", "step_index_counter"])
        incl_columns = [
            "current",
            "voltage",
            "charge_capacity",
            "discharge_capacity",
            "charge_energy",
            "discharge_energy",
            "internal_resistance",
            "temperature",
            "test_time",
        ]

        diag_dict = {}
        for cycle in diag_data.cycle_index.unique():
            diag_dict.update({cycle: None})
            steps = diag_data[diag_data.cycle_index == cycle].step_index.unique()
            diag_dict[cycle] = list(steps)

        all_dfs = []
        for (cycle_index, step_index, step_index_counter), df in tqdm(group):
            if len(df) < 2:
                continue
            step_voltage_delta = df.voltage.max() - df.voltage.min()
            if diag_cycle_type[diag_cycles_at.index(cycle_index)] == "hppc" and step_voltage_delta >= v_delta_min:
                v_hppc_step = [df.voltage.min(), df.voltage.max()]
                hppc_resolution = int(
                    (df.voltage.max() - df.voltage.min()) / v_resolution
                )
                new_df = interpolate_df(
                    df,
                    field_name="voltage",
                    field_range=v_hppc_step,
                    columns=incl_columns,
                    resolution=hppc_resolution,
                )
            elif step_voltage_delta < v_delta_min:
                t_range_step = [df.test_time.min(), df.test_time.max()]
                new_df = interpolate_df(
                    df,
                    field_name="test_time",
                    field_range=t_range_step,
                    columns=incl_columns,
                    resolution=resolution,
                )
            else:
                new_df = interpolate_df(
                    df,
                    field_name="voltage",
                    field_range=v_range,
                    columns=incl_columns,
                    resolution=resolution,
                )

            new_df["cycle_index"] = cycle_index
            new_df["cycle_type"] = diag_cycle_type[diag_cycles_at.index(cycle_index)]
            new_df["step_index"] = step_index
            new_df["step_index_counter"] = step_index_counter
            new_df["step_type"] = diag_dict[cycle_index].index(step_index)
            new_df.astype(
                {
                    "cycle_index": "int32",
                    "cycle_type": "category",
                    "step_index": "uint8",
                    "step_index_counter": "int16",
                    "step_type": "uint8",
                }
            )
            new_df["discharge_dQdV"] = (
                new_df.discharge_capacity.diff() / new_df.voltage.diff()
            )
            new_df["charge_dQdV"] = (
                new_df.charge_capacity.diff() / new_df.voltage.diff()
            )
            all_dfs.append(new_df)

        # Ignore the index to avoid issues with overlapping voltages
        result = pd.concat(all_dfs, ignore_index=True)
        # Cycle_index gets a little weird about typing, so round it here
        result.cycle_index = result.cycle_index.round()
        result = self._cast_dtypes(result, "diagnostic_interpolated")

        return result

    def summarize_diagnostic(self, diagnostic_available):
        """
        Gets summary statistics for data according to location of
        diagnostic cycles in the data

        Args:
            diagnostic_available (dict): dictionary with diagnostic_types
                as list, 'length' of the diagnostic in cycles and location
                of the diagnostic by cycle index

        Returns:
            (DataFrame) of summary statistics by cycle

        """

        max_cycle = self.raw_data.cycle_index.max()
        starts_at = [
            i for i in diagnostic_available["diagnostic_starts_at"] if i <= max_cycle
        ]
        diag_cycles_at = list(
            itertools.chain.from_iterable(
                [list(range(i, i + diagnostic_available["length"])) for i in starts_at]
            )
        )
        diag_summary = self.raw_data.groupby("cycle_index").agg(self._diag_aggregation)

        diag_summary.columns = self._diag_summary_cols

        diag_summary = diag_summary[diag_summary.index.isin(diag_cycles_at)]

        diag_summary["coulombic_efficiency"] = (
            diag_summary["discharge_capacity"] / diag_summary["charge_capacity"]
        )
        diag_summary["paused"] = self.raw_data.groupby("cycle_index").apply(
            get_max_paused_over_threshold
        )

        diag_summary.reset_index(drop=True, inplace=True)

        diag_summary["cycle_type"] = pd.Series(
            diagnostic_available["cycle_type"] * len(starts_at)
        )

        # Add CV_time, CV_current, and CV_capacity summary stats
        CV_time = []
        CV_current = []
        CV_capacity = []
        for cycle in diag_summary.cycle_index:
            raw_cycle = self.raw_data.loc[self.raw_data.cycle_index == cycle]

            # Charge is the very first step_index
            CCCV = raw_cycle.loc[raw_cycle.step_index == raw_cycle.step_index.min()]
            CV = get_CV_segment_from_charge(CCCV)
            if CV.empty:
                logger.debug(f"Failed to extract CV segment for diagnostic cycle {cycle}!")
            CV_time.append(get_CV_time(CV))
            CV_current.append(get_CV_current(CV))
            CV_capacity.append(get_CV_capacity(CV))

        diag_summary["CV_time"] = CV_time
        diag_summary["CV_current"] = CV_current
        diag_summary["CV_capacity"] = CV_capacity

        diag_summary = self._cast_dtypes(diag_summary, "diagnostic_summary")

        return diag_summary

    # locate diagnostic cycles
    # determine voltage range to interpolate on
    # this is a function that TRI is using mostly for themselves

    @StructuringDecorators.must_not_be_legacy
    def determine_structuring_parameters(
        self,
        v_range=None,
        resolution=1000,
        nominal_capacity=1.1,
        full_fast_charge=0.8,
        parameters_path=PROTOCOL_PARAMETERS_DIR,
    ):
        """
        Method for determining what values to use to convert raw run into processed run.


        Args:
            v_range ([float, float]): voltage range for interpolation
            resolution (int): resolution for interpolation
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            parameters_path (str): path to parameters files

        Returns:
            v_range ([float, float]): voltage range for interpolation
            resolution (int): resolution for interpolation
            nominal_capacity (float): nominal capacity for summary stats
            full_fast_charge (float): full fast charge for summary stats
            diagnostic_available (dict): dictionary of values to use for
                finding and using the diagnostic cycles

        """
        if not parameters_path or not os.path.exists(parameters_path):
            raise FileNotFoundError(f"Parameters path {parameters_path} does not exist!")

        run_parameter, all_parameters = parameters_lookup.get_protocol_parameters(
            self.paths["raw"], parameters_path
        )

        # Logic for interpolation variables and diagnostic cycles
        diagnostic_available = False
        if run_parameter is not None:
            if {"capacity_nominal"}.issubset(run_parameter.columns.tolist()):
                nominal_capacity = run_parameter["capacity_nominal"].iloc[0]
            if {"discharge_cutoff_voltage", "charge_cutoff_voltage"}.issubset(
                    run_parameter.columns):
                v_range = [
                    all_parameters["discharge_cutoff_voltage"].min(),
                    all_parameters["charge_cutoff_voltage"].max(),
                ]
            if {"diagnostic_type", "diagnostic_start_cycle", "diagnostic_interval"}.issubset(run_parameter.columns):
                if run_parameter["diagnostic_type"].iloc[0] == "HPPC+RPT":
                    hppc_rpt = ["reset", "hppc", "rpt_0.2C", "rpt_1C", "rpt_2C"]
                    hppc_rpt_len = 5
                    initial_diagnostic_at = [1, 1 + run_parameter["diagnostic_start_cycle"].iloc[0] + 1 * hppc_rpt_len]
                    # Calculate the number of steps present for each cycle in the diagnostic as the pattern for
                    # the diagnostic. If this pattern of steps shows up at the end of the file, that indicates
                    # the presence of a final diagnostic
                    diag_0_pattern = [len(self.raw_data[self.raw_data.cycle_index == x].step_index.unique())
                                      for x in range(initial_diagnostic_at[0], initial_diagnostic_at[0] + hppc_rpt_len)]
                    diag_1_pattern = [len(self.raw_data[self.raw_data.cycle_index == x].step_index.unique())
                                      for x in range(initial_diagnostic_at[1], initial_diagnostic_at[1] + hppc_rpt_len)]
                    # Find the steps present in the reset cycles for the first and second diagnostic
                    diag_0_steps = set(self.raw_data[self.raw_data.cycle_index ==
                                       initial_diagnostic_at[0]].step_index.unique())
                    diag_1_steps = set(self.raw_data[self.raw_data.cycle_index ==
                                       initial_diagnostic_at[1]].step_index.unique())
                    diagnostic_starts_at = []
                    for cycle in self.raw_data.cycle_index.unique():
                        steps_present = set(self.raw_data[
                                                self.raw_data.cycle_index == cycle].step_index.unique())
                        cycle_pattern = [len(self.raw_data[
                                                 self.raw_data.cycle_index == x].step_index.unique())
                                         for x in
                                         range(cycle, cycle + hppc_rpt_len)]
                        if steps_present == diag_0_steps or steps_present == diag_1_steps:
                            diagnostic_starts_at.append(cycle)
                        # Detect final diagnostic if present in the data
                        elif cycle >= (
                                self.raw_data.cycle_index.max() - hppc_rpt_len - 1) and \
                                (cycle_pattern == diag_0_pattern or cycle_pattern == diag_1_pattern):
                            diagnostic_starts_at.append(cycle)

                    diagnostic_available = {
                        "parameter_set":
                            run_parameter["diagnostic_parameter_set"].iloc[0],
                        "cycle_type": hppc_rpt,
                        "length": hppc_rpt_len,
                        "diagnostic_starts_at": diagnostic_starts_at,
                    }

        return (
            v_range,
            resolution,
            nominal_capacity,
            full_fast_charge,
            diagnostic_available,
        )

    @StructuringDecorators.must_be_structured
    def get_cycle_life(self, n_cycles_cutoff=40, threshold=0.8):
        """
        Calculate cycle life for capacity loss below a certain threshold

        Args:
            n_cycles_cutoff (int): cutoff for number of cycles to sample
                for the cycle life in order to use median method.
            threshold (float): fraction of capacity loss for which
                to find the cycle index.

        Returns:
            float: cycle life.
        """
        # discharge_capacity has a spike and  then increases slightly between \
        # 1-40 cycles, so let us use take median of 1st 40 cycles for max.

        # If n_cycles <  n_cycles_cutoff, do not use median method
        if len(self.structured_summary) > n_cycles_cutoff:
            max_capacity = np.median(
                self.structured_summary.discharge_capacity.iloc[0:n_cycles_cutoff]
            )
        else:
            max_capacity = 1.1

        # If capacity falls below 80% of initial capacity by end of run
        if self.structured_summary.discharge_capacity.iloc[-1] / max_capacity <= threshold:
            cycle_life = self.structured_summary[
                self.structured_summary.discharge_capacity < threshold * max_capacity
            ].index[0]
        else:
            # Some cells do not degrade below the threshold (low degradation rate)
            cycle_life = len(self.structured_summary) + 1

        return cycle_life

    @StructuringDecorators.must_be_structured
    def cycles_to_capacities(self, cycle_min=200, cycle_max=1800, cycle_interval=200):
        """
        Get discharge capacity at constant intervals of 200 cycles

        Args:
            cycle_min (int): Cycle number to being forecasting capacity at
            cycle_max (int): Cycle number to end forecasting capacity at
            cycle_interval (int): Intervals for forecasts

        Returns:
            pandas.DataFrame:
        """
        discharge_capacities = pd.DataFrame(
            np.zeros((1, int((cycle_max - cycle_min) / cycle_interval)))
        )
        counter = 0
        cycle_indices = np.arange(cycle_min, cycle_max, cycle_interval)
        for cycle_index in cycle_indices:
            try:
                discharge_capacities[counter] = self.structured_summary.discharge_capacity.iloc[
                    cycle_index
                ]
            except IndexError:
                pass
            counter = counter + 1
        discharge_capacities.columns = np.core.defchararray.add(
            "cycle_", cycle_indices.astype(str)
        )
        return discharge_capacities

    @StructuringDecorators.must_be_structured
    def capacities_to_cycles(
        self, thresh_max_cap=0.98, thresh_min_cap=0.78, interval_cap=0.03
    ):
        """
        Get cycles to reach set threshold capacities.

        Args:
            thresh_max_cap (float): Upper bound on capacity to compute cycles at.
            thresh_min_cap (float): Lower bound on capacity to compute cycles at.
            interval_cap (float): Interval/step size.

        Returns:
            pandas.DataFrame:
        """
        threshold_list = np.around(
            np.arange(thresh_max_cap, thresh_min_cap, -interval_cap), 2
        )
        counter = 0
        cycles = pd.DataFrame(np.zeros((1, len(threshold_list))))
        for threshold in threshold_list:
            cycles[counter] = self.get_cycle_life(threshold=threshold)
            counter = counter + 1
        cycles.columns = np.core.defchararray.add(
            "capacity_", threshold_list.astype(str)
        )
        return cycles

    @property
    def paused_intervals(self):
        # a method to use get_max_paused_over_threshold
        return self.raw_data.groupby("cycle_index").apply(get_max_paused_over_threshold)

    @property
    def is_structured(self):
        required = [self.structured_summary, self.structured_data, self.diagnostic_summary, self.diagnostic_data]
        if any([df is not None for df in required]):
            return True
        else:
            return False

    def _cast_dtypes(self, result, structure_dtypes_key):
        """Cast data types of a result dataframe to those specified by the structuring config.

        Args:
            result (pd.DataFrame): The result to cast
            structure_dtypes_key: The required key, for structure dtype casting, as per STRUCTURE_DTYPES

        Returns:
            (pd.DataFrame): The result, cast to the correct datatypes.

        """
        available_dtypes = {}
        for field, dtype in STRUCTURE_DTYPES[structure_dtypes_key].items():
            if field in result.columns:
                # if not result[field].isna().all():
                available_dtypes[field] = dtype

        return result.astype(available_dtypes)


# based on get_interpolated_data
def interpolate_df(
        dataframe,
        field_name="voltage",
        field_range=None,
        columns=None,
        resolution=1000
):
    """
    General method for creating uniform (i. e. same # of data points)
    dataframe using an by interpolation of supplied columns on the
    specified field, assumes input field data is monotonic

    Args:
        dataframe (pandas.DataFrame): dataframe to create interpolate
        field_name (str): column name to use as the dependent interpolation variable
        field_range (list): list of two values to use as endpoints, if None,
            range is the min/max of the dataframe field_name values
        columns (list): list of column names to provide interpolated values for,
            default value of None indicates all columns should be interpolated
        resolution (int): number of data points to sample in the uniform
            cycles, defaults to 1000

    Returns:
        pandas.DataFrame: DataFrame of interpolated values
    """
    columns = columns or dataframe.columns
    columns = list(set(columns) | {field_name})

    df = dataframe.loc[:, dataframe.columns.intersection(columns)]
    field_range = field_range or [df[field_name].iloc[0],
                                  df[field_name].iloc[-1]]
    # If interpolating on datetime, change column to datetime series and
    # use date_range to create interpolating vector
    if field_name == "date_time_iso":
        df["date_time_iso"] = pd.to_datetime(df["date_time_iso"])
        interp_x = pd.date_range(
            start=df[field_name].iloc[0],
            end=df[field_name].iloc[-1],
            periods=resolution,
        )
    else:
        interp_x = np.linspace(*field_range, resolution)
    interpolated_df = pd.DataFrame({field_name: interp_x, "interpolated": True})

    df["interpolated"] = False

    # Merge interpolated and uninterpolated DFs to use pandas interpolation
    interpolated_df = interpolated_df.merge(df, how="outer", on=field_name, sort=True)
    interpolated_df = interpolated_df.set_index(field_name)

    interpolated_df = interpolated_df.interpolate("slinear")

    # Filter for only interpolated values
    interpolated_df[["interpolated_x"]] = interpolated_df[
        ["interpolated_x"]].fillna(
        False
    )
    interpolated_df = interpolated_df[interpolated_df["interpolated_x"]]
    interpolated_df = interpolated_df.drop(["interpolated_x", "interpolated_y"],
                                           axis=1)
    interpolated_df = interpolated_df.reset_index()
    # Remove duplicates
    interpolated_df = interpolated_df[~interpolated_df[field_name].duplicated()]
    return interpolated_df


def step_is_chg_state(step_df, chg):
    """
    Helper function to determine whether a given dataframe corresponding
    to a single cycle_index/step is charging or discharging, only intended
    to be used with a dataframe for single step/cycle

    Args:
         step_df (pandas.DataFrame): dataframe to determine whether
            charging or discharging
         chg (bool): Charge state; True if charging

    Returns:
        (bool): True if step is the charge state specified.
    """
    cap = step_df[["charge_capacity", "discharge_capacity"]]
    cap = cap.diff(axis=0).mean(axis=0).diff().iloc[-1]

    if chg:  # Charging
        return cap < 0
    else:  # Discharging
        return cap > 0


def step_is_dchg(step_df):
    return step_is_chg_state(step_df, False)


def step_is_chg(step_df):
    return step_is_chg_state(step_df, True)


def step_is_waveform(step_df, chg_filter):
    """
    Helper function for driving profiles to determine whether a given dataframe corresponding
    to a single cycle_index/step is a waveform discharge.

    Args:
         step_df (pandas.DataFrame): dataframe to determine whether waveform step is present
         chg_filter (func): Function for determining whether step is charging or not.


    Returns:
        (bool): True if step is waveform discharge.
    """

    # Check for waveform in maccor

    voltage_resolution = 3
    if len([col for col in step_df.columns if '_wf_' in col]):
        return (chg_filter(step_df)) & \
               ((step_df['_wf_chg_cap'].notna().any()) |
                (step_df['_wf_dis_cap'].notna().any()))
    elif not np.round(step_df.voltage, voltage_resolution).is_monotonic:
        # This is a placeholder logic for arbin waveform detection
        # This fails for some arbin files that nominally have a CC-CV step.
        # e.g. 2017-12-04_4_65C-69per_6C_CH29.csv
        # TODO: survey more files and include additional heuristics/logic based on the size of
        # and frequency of non-monotonicities to determine whether step is actually a waveform.

        # All non-maccor files will evaluate to False by default for now
        return False
    else:
        return False


def step_is_waveform_dchg(step_df):
    return step_is_waveform(step_df, step_is_dchg)


def step_is_waveform_chg(step_df):
    return step_is_waveform(step_df, step_is_chg)


def get_max_paused_over_threshold(group, paused_threshold=3600):
    """
    Evaluate a raw cycling dataframe to determine if there is a pause in cycling.
    The method looks at the time difference between each row and if this value
    exceeds a threshold, it returns that length of time in seconds. Otherwise it
    returns 0

    Args:
        group (pd.DataFrame): cycling dataframe with date_time_iso column
        paused_threshold (int): gap in seconds to classify as a pause in cycling

    Returns:
        (float): number of seconds that test was paused

    """
    date_time_objs = pd.to_datetime(group["date_time_iso"])
    date_time_float = [
        time.mktime(t.timetuple()) if t is not pd.NaT else float("nan")
        for t in date_time_objs
    ]
    date_time_float = pd.Series(date_time_float)
    if date_time_float.diff().max() > paused_threshold:
        max_paused_duration = date_time_float.diff().max()
    else:
        max_paused_duration = 0
    return max_paused_duration


def get_CV_segment_from_charge(charge, dt_tol=1, dVdt_tol=1e-5, dIdt_tol=1e-4):
    """
    Extracts the constant voltage segment from charge. Works for both CCCV or
    CC steps followed by a CV step.

    Args:
        charge (pd.DataFrame): charge dataframe for a single cycle
        dt_tol (float) : dt tolernace (minimum) for identifying CV
        dVdt_tol (float) : dVdt tolerance (maximum) for identifying CV
        dIdt_tol (float) : dVdt tolerance (minimum) for identifying CV 

    Returns:
        (pd.DataFrame): dataframe containing the CV segment

    """
    if charge.empty:
        return(charge)
    else:
        # Compute dI and dV
        dI = np.diff(charge.current)
        dV = np.diff(charge.voltage)
        dt = np.diff(charge.test_time)

        # Find the first index where the CV segment begins
        i = 0
        while i < len(dV) and (dt[i] < dt_tol or abs(dV[i]/dt[i]) > dVdt_tol or abs(dI[i]/dt[i]) < dIdt_tol):
            i = i+1

        # Filter for CV phase
        return(charge.loc[charge.test_time >= charge.test_time.iat[i-1]])


def get_CV_time(CV):
    """
    Helper function to compute CV time.

    Args:
        CV (pd.DataFrame): CV segement of charge

    Returns:
        (float): length of the CV segment in seconds

    """
    if not CV.empty:
        return(CV.test_time.iat[-1] - CV.test_time.iat[0])


def get_CV_current(CV):
    """
    Helper function to compute CV current.

    Args:
        CV (pd.DataFrame): CV segement of charge

    Returns:
        (float): current reached at the end of the CV segment

    """
    if not CV.empty:
        return(CV.current.iat[-1])


def get_CV_capacity(CV):
    """
    Helper function to compute CV capacity.

    Args:
        CV (pd.DataFrame): CV segement of charge

    Returns:
        (float): charge capacity during the CV segment

    """
    if not CV.empty:
        return(CV.charge_capacity.iat[-1] - CV.charge_capacity.iat[0])



#  Copyright (c) 2019 Toyota Research Institute
import os

PROTOCOL_DIR = os.path.dirname(__file__)
PROCEDURE_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "procedure_templates")
SCHEDULE_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "schedule_templates")
PROTOCOL_SCHEMA_DIR = os.path.join(PROTOCOL_DIR, "protocol_schemas")
BIOLOGIC_TEMPLATE_DIR = os.path.join(PROTOCOL_DIR, "biologic_templates")



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to cycler run data structures"""

import os
import copy
import unittest

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from monty.serialization import loadfn, dumpfn
from monty.tempfile import ScratchDir

from beep.conversion_schemas import STRUCTURE_DTYPES
from beep.utils.s3 import download_s3_object
from beep.structure.base import BEEPDatapath
from beep.structure.base_eis import EIS, BEEPDatapathWithEIS
from beep.structure.maccor import MaccorDatapath
from beep.tests.constants import TEST_FILE_DIR, BIG_FILE_TESTS, SKIP_MSG
from beep.structure.cli import auto_load_processed
from beep import VALIDATION_SCHEMA_DIR


class TestBEEPDatapath(unittest.TestCase):
    """
    Tests common to all datapaths.
    """

    class BEEPDatapathChildTest(BEEPDatapath):
        """
        A test class representing any child of BEEPDatapath.
        """

        @classmethod
        def from_file(cls, path, index_col=0, **kwargs):
            return pd.read_csv(path, index_col=0, **kwargs)

    @classmethod
    def setUpClass(cls) -> None:

        # Use arbin memloaded inputs as source of non-diagnostic truth
        arbin_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_arbin_memloaded.csv")
        arbin_meta_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_arbin_metadata_memloaded.json")
        cls.data_nodiag = pd.read_csv(arbin_fname, index_col=0)
        cls.metadata_nodiag = loadfn(arbin_meta_fname)
        cls.datapath_nodiag = cls.BEEPDatapathChildTest(
            raw_data=cls.data_nodiag,
            metadata=cls.metadata_nodiag,
            paths={"raw": arbin_fname, "raw_metadata": arbin_meta_fname}
        )

        # Use maccor memloaded inputs as source of diagnostic truth
        maccor_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_w_diagnostic_memloaded.csv")
        maccor_meta_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_w_diagnostic_metadata_memloaded.json")
        cls.data_diag = pd.read_csv(maccor_fname, index_col=0)
        cls.metadata_diag = loadfn(maccor_meta_fname)
        cls.datapath_diag = cls.BEEPDatapathChildTest(
            raw_data=cls.data_diag,
            metadata=cls.metadata_diag,
            paths={"raw": maccor_fname, "raw_metadata": maccor_meta_fname}

        )

        # Use maccor paused memloaded inputs as source of paused run truth
        maccor_paused_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_paused_memloaded.csv")
        maccor_paused_meta_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_paused_metadata_memloaded.json")
        cls.data_paused = pd.read_csv(maccor_paused_fname, index_col=0)
        cls.metadata_paused = loadfn(maccor_paused_meta_fname)
        cls.datapath_paused = cls.BEEPDatapathChildTest(
            raw_data=cls.data_paused,
            metadata=cls.metadata_paused,
            paths={"raw": maccor_paused_fname, "raw_metadata": maccor_paused_meta_fname}
        )

        # Small maccor file with parameters
        maccor_small_params_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_parameterized_memloaded.csv")
        maccor_small_params_meta_fname = os.path.join(TEST_FILE_DIR,
                                                      "BEEPDatapath_maccor_parameterized_metadata_memloaded.json")
        cls.data_small_params = pd.read_csv(maccor_small_params_fname, index_col=0)
        cls.metadata_small_params = loadfn(maccor_small_params_meta_fname)
        cls.datapath_small_params = cls.BEEPDatapathChildTest(
            raw_data=cls.data_small_params,
            metadata=cls.metadata_small_params,
            paths={"raw": maccor_small_params_fname, "raw_metadata": maccor_small_params_meta_fname}
        )

        # Maccor with various diagnostics from memory
        # For testing determine_structuring_parameters
        maccor_diag_normal_fname = os.path.join(TEST_FILE_DIR, "BEEPDatapath_maccor_diagnostic_normal_memloaded.csv")
        maccor_diag_normal_meta_fname = os.path.join(TEST_FILE_DIR,
                                                     "BEEPDatapath_maccor_diagnostic_normal_metadata_memloaded.json")
        maccor_diag_normal_original_fname = os.path.join(TEST_FILE_DIR, "PreDiag_000287_000128short.092")
        cls.data_diag_normal = pd.read_csv(maccor_diag_normal_fname, index_col=0)
        cls.metadata_diag_normal = loadfn(maccor_diag_normal_meta_fname)
        cls.datapath_diag_normal = cls.BEEPDatapathChildTest(
            raw_data=cls.data_diag_normal,
            metadata=cls.metadata_diag_normal,
            paths={"raw": maccor_diag_normal_original_fname, "raw_metadata": maccor_diag_normal_meta_fname}
        )
        maccor_diag_misplaced_fname = os.path.join(TEST_FILE_DIR,
                                                   "BEEPDatapath_maccor_diagnostic_misplaced_memloaded.csv")
        maccor_diag_misplaced_meta_fname = os.path.join(
            TEST_FILE_DIR, "BEEPDatapath_maccor_diagnostic_misplaced_metadata_memloaded.json"
        )
        maccor_diag_misplaced_original_fname = os.path.join(TEST_FILE_DIR, "PreDiag_000412_00008Fshort.022")
        cls.data_diag_misplaced = pd.read_csv(maccor_diag_misplaced_fname, index_col=0)
        cls.metadata_diag_misplaced = loadfn(maccor_diag_misplaced_meta_fname)
        cls.datapath_diag_misplaced = cls.BEEPDatapathChildTest(
            raw_data=cls.data_diag_misplaced,
            metadata=cls.metadata_diag_misplaced,
            paths={"raw": maccor_diag_misplaced_original_fname, "raw_metadata": maccor_diag_misplaced_meta_fname}
        )

        cls.cycle_run_file = os.path.join(
            TEST_FILE_DIR, "2017-12-04_4_65C-69per_6C_CH29_processed.json"
        )

        cls.diagnostic_available = {
            "type": "HPPC",
            "cycle_type": ["hppc"],
            "length": 1,
            "diagnostic_starts_at": [1],
        }

    def setUp(self) -> None:
        # Reset all datapaths after each run, avoiding colliions between tests
        # and avoiding reloading files
        for dp in [
            self.datapath_diag,
            self.datapath_nodiag,
            self.datapath_diag_normal,
            self.datapath_diag_misplaced,
            self.datapath_paused,
            self.datapath_small_params
        ]:
            dp.unstructure()

    def run_dtypes_check(self, summary):
        reg_dyptes = summary.dtypes.tolist()
        reg_columns = summary.columns.tolist()
        reg_dyptes = [str(dtyp) for dtyp in reg_dyptes]
        for indx, col in enumerate(reg_columns):
            self.assertEqual(reg_dyptes[indx], STRUCTURE_DTYPES["summary"][col])

    def test_abc(self):

        class BEEPDatapathChildBad(BEEPDatapath):
            def some_extra_method(self):
                return True

            # missing required method!
            # should fail

        # Ensure this bad child class has ABC error
        with self.assertRaises(TypeError):
            BEEPDatapathChildBad(raw_data=pd.DataFrame({}), metadata={"some": "metadata"})

    def test_unstructure(self):
        self.datapath_nodiag.structure()
        self.assertTrue(self.datapath_nodiag.is_structured)
        self.datapath_nodiag.unstructure()
        self.assertFalse(self.datapath_nodiag.is_structured)

    def test_serialization(self):
        truth_datapath = self.datapath_diag

        truth_datapath.structure()
        d = truth_datapath.as_dict()
        datapath_from_dict = self.BEEPDatapathChildTest.from_dict(d)

        # Test loading with and without compression, and with and without raw_data via omit_raw
        for fname_short in ("test_serialization.json", "test_serialization.json.gz"):
            for omit_raw in (True, False):

                fname = os.path.join(TEST_FILE_DIR, fname_short)
                truth_datapath.to_json_file(fname, omit_raw=omit_raw)
                datapath_from_json = self.BEEPDatapathChildTest.from_json_file(fname)

                for df_name in ("structured_data", "structured_summary", "diagnostic_data", "diagnostic_summary"):
                    df_truth = getattr(truth_datapath, df_name)
                    for datapath_test in (datapath_from_dict, datapath_from_json):
                        df_test = getattr(datapath_test, df_name)

                        if df_truth is None:
                            self.assertEqual(df_truth, df_test)
                        else:
                            self.assertTrue(isinstance(df_test, pd.DataFrame))
                            self.assertTrue(df_truth.equals(df_test))

                self.assertEqual(datapath_from_json.paths.get("structured"), fname)
                self.assertEqual(datapath_from_json.paths.get("raw"), self.datapath_diag.paths.get("raw"))

                if os.path.exists(fname):
                    os.remove(fname)

    # Test to address bug where schema_path is absolute but is created in a different environment
    def test_reloading_new(self):
        test_file = os.path.join(TEST_FILE_DIR, "PredictionDiagnostics_000107_0001B9_structure_short.json")
        struct = auto_load_processed(test_file)
        self.assertEqual(struct.schema,
                         os.path.join(VALIDATION_SCHEMA_DIR, "schema-maccor-2170.yaml"))

    # based on RCRT.test_serialization
    def test_serialization_legacy(self):
        test_file = os.path.join(
            TEST_FILE_DIR, "2017-12-04_4_65C-69per_6C_CH29_processed.json"
        )

        datapath = self.BEEPDatapathChildTest.from_json_file(test_file)
        self.assertTrue(isinstance(datapath.structured_summary, pd.DataFrame))
        self.assertTrue(isinstance(datapath.structured_data, pd.DataFrame))
        self.assertIsNone(datapath.metadata.barcode)
        self.assertIsNone(datapath.metadata.protocol)
        self.assertEqual(datapath.metadata.channel_id, 28)
        self.assertEqual(datapath.structured_summary.shape[0], 188)
        self.assertEqual(datapath.structured_summary.shape[1], 7)
        self.assertEqual(datapath.structured_data.shape[0], 188000),
        self.assertEqual(datapath.structured_data.shape[1], 7)
        self.assertAlmostEqual(datapath.structured_data["voltage"].loc[0], 2.8, places=5)
        self.assertAlmostEqual(datapath.structured_data["discharge_capacity"].loc[187999], 0.000083, places=6)
        self.assertEqual(datapath.paths.get("structured"), test_file)
        self.assertEqual(datapath.paths.get("raw"), None)

    # based on RCRT.test_get_interpolated_charge_step
    def test_interpolate_step(self):
        reg_cycles = [i for i in self.datapath_nodiag.raw_data.cycle_index.unique()]
        v_range = [2.8, 3.5]
        resolution = 1000

        for step_type in ("charge", "discharge"):
            interpolated_charge = self.datapath_nodiag.interpolate_step(
                v_range,
                resolution,
                step_type=step_type,
                reg_cycles=reg_cycles,
                axis="test_time",
            )
            lengths = [len(df) for index, df in interpolated_charge.groupby("cycle_index")]
            axis_1 = interpolated_charge[
                interpolated_charge.cycle_index == 5
            ].charge_capacity.to_list()
            axis_2 = interpolated_charge[
                interpolated_charge.cycle_index == 10
            ].charge_capacity.to_list()
            self.assertGreater(max(axis_1), max(axis_2))
            self.assertTrue(np.all(np.array(lengths) == 1000))

            if step_type == "charge":
                self.assertTrue(interpolated_charge["current"].mean() > 0)
            else:
                self.assertTrue(interpolated_charge["current"].mean() < 0)

    # based on RCRT.test_get_interpolated_discharge_cycles
    # based on RCRT.test_get_interpolated_charge_cycles
    # based on RCRT.test_interpolated_cycles_dtypes
    def test_interpolate_cycles(self):
        dp = self.datapath_nodiag
        all_interpolated = dp.interpolate_cycles()

        # Test discharge cycles
        dchg = all_interpolated[(all_interpolated.step_type == "discharge")]
        lengths = [len(df) for index, df in dchg.groupby("cycle_index")]
        self.assertTrue(np.all(np.array(lengths) == 1000))

        # Found these manually
        dchg = dchg.drop(columns=["step_type"])
        y_at_point = dchg.iloc[[1500]]
        x_at_point = dchg.voltage[1500]
        cycle_1 = dp.raw_data[dp.raw_data["cycle_index"] == 1]

        # Discharge step is 12
        discharge = cycle_1[cycle_1.step_index == 12]
        discharge = discharge.sort_values("voltage")

        # Get an interval between which one can find the interpolated value
        measurement_index = np.max(np.where(discharge.voltage - x_at_point < 0))
        interval = discharge.iloc[measurement_index: measurement_index + 2]
        interval = interval.drop(columns=["date_time_iso"])  # Drop non-numeric column

        # Test interpolation with a by-hand calculation of slope
        diff = np.diff(interval, axis=0)
        pred = interval.iloc[[0]] + diff * (
                    x_at_point - interval.voltage.iloc[0]) / (
                       interval.voltage.iloc[1] - interval.voltage.iloc[0]
               )
        pred = pred.reset_index()
        for col_name in y_at_point.columns:
            self.assertAlmostEqual(
                pred[col_name].iloc[0], y_at_point[col_name].iloc[0],
                places=2
            )

        # Test charge cycles
        chg = all_interpolated[(all_interpolated.step_type == "charge")]
        lengths = [len(df) for index, df in chg.groupby("cycle_index")]
        axis_1 = chg[chg.cycle_index == 5].charge_capacity.to_list()
        axis_2 = chg[chg.cycle_index == 10].charge_capacity.to_list()
        self.assertEqual(axis_1, axis_2)
        self.assertTrue(np.all(np.array(lengths) == 1000))
        self.assertTrue(chg["current"].mean() > 0)


        # Test dtypes
        cycles_interpolated_dyptes = all_interpolated.dtypes.tolist()
        cycles_interpolated_columns = all_interpolated.columns.tolist()
        cycles_interpolated_dyptes = [str(dtyp) for dtyp in cycles_interpolated_dyptes]
        for indx, col in enumerate(cycles_interpolated_columns):
            self.assertEqual(
                cycles_interpolated_dyptes[indx],
                STRUCTURE_DTYPES["cycles_interpolated"][col],
            )

    # based on RCRT.test_get_summary
    # based on RCRT.test_get_energy
    # based on RCRT.test_get_charge_throughput
    # based on RCRT.test_summary_dtypes
    def test_summarize_cycles(self):
        summary_diag = self.datapath_diag.summarize_cycles(nominal_capacity=4.7, full_fast_charge=0.8)
        self.assertTrue(
            set.issubset(
                {
                    "discharge_capacity",
                    "charge_capacity",
                    "dc_internal_resistance",
                    "temperature_maximum",
                    "temperature_average",
                    "temperature_minimum",
                    "date_time_iso",
                    "charge_throughput",
                    "energy_throughput",
                    "charge_energy",
                    "discharge_energy",
                    "energy_efficiency",
                    "CV_time",
                    "CV_current",
                    "CV_capacity"
                },
                set(summary_diag.columns),
            )
        )
        self.assertEqual(summary_diag["cycle_index"].tolist(), list(range(0, 13)))
        self.assertEqual(len(summary_diag.index), len(summary_diag["date_time_iso"]))
        self.assertEqual(summary_diag["paused"].max(), 0)
        self.assertEqual(summary_diag["CV_time"][1], np.float32(160111.796875))
        self.assertEqual(summary_diag["CV_current"][1], np.float32(0.4699016))
        self.assertEqual(summary_diag["CV_capacity"][1], np.float32(94.090355))
        self.run_dtypes_check(summary_diag)

        # incorporates test_get_energy and get_charge_throughput
        summary = self.datapath_nodiag.summarize_cycles(nominal_capacity=4.7, full_fast_charge=0.8)
        self.assertEqual(np.around(summary["charge_energy"][5], 6), np.around(3.7134638, 6))
        self.assertEqual(np.around(summary["energy_efficiency"][5], 7), np.around(np.float32(0.872866405753033), 7))
        self.assertEqual(summary["charge_throughput"][5], np.float32(6.7614093))
        self.assertEqual(summary["energy_throughput"][5], np.float32(23.2752363))
        self.run_dtypes_check(summary)

    # based on RCRT.test_determine_structuring_parameters
    def test_determine_structuring_parameters(self):
        (v_range, resolution, nominal_capacity, full_fast_charge, diagnostic_available) = \
            self.datapath_diag_normal.determine_structuring_parameters()
        diagnostic_available_test = {
            "parameter_set": "Tesla21700",
            "cycle_type": ["reset", "hppc", "rpt_0.2C", "rpt_1C", "rpt_2C"],
            "length": 5,
            "diagnostic_starts_at": [
                1, 36, 141, 246, 351, 456, 561, 666, 771, 876, 981, 1086,
                1191,
                1296, 1401, 1506, 1611, 1716, 1821, 1926, 2031, 2136, 2241,
                2346,
                2451, 2556, 2661, 2766, 2871, 2976, 3081, 3186, 3291, 3396,
                3501,
                3606, 3628
            ]
        }

        self.assertEqual(v_range, [2.5, 4.2])
        self.assertEqual(resolution, 1000)
        self.assertEqual(nominal_capacity, 4.84)
        self.assertEqual(full_fast_charge, 0.8)
        self.assertEqual(diagnostic_available, diagnostic_available_test)

        (
            v_range,
            resolution,
            nominal_capacity,
            full_fast_charge,
            diagnostic_available,
        ) = self.datapath_diag_misplaced.determine_structuring_parameters()
        diagnostic_available_test = {
            "parameter_set": "Tesla21700",
            "cycle_type": ["reset", "hppc", "rpt_0.2C", "rpt_1C", "rpt_2C"],
            "length": 5,
            "diagnostic_starts_at": [1, 36, 141, 220, 255]
        }
        self.assertEqual(v_range, [2.5, 4.2])
        self.assertEqual(resolution, 1000)
        self.assertEqual(nominal_capacity, 4.84)
        self.assertEqual(full_fast_charge, 0.8)
        self.assertEqual(diagnostic_available, diagnostic_available_test)

    # based on RCRT.test_get_interpolated_diagnostic_cycles
    def test_interpolate_diagnostic_cycles(self):
        d_interp = self.datapath_diag.interpolate_diagnostic_cycles(
            self.diagnostic_available, resolution=500
        )
        self.assertGreaterEqual(len(d_interp.cycle_index.unique()), 1)

        # Ensure step indices are partitioned and processed separately
        self.assertEqual(len(d_interp.step_index.unique()), 9)
        first_step = d_interp[
            (d_interp.step_index == 7) & (d_interp.step_index_counter == 1)
        ]
        second_step = d_interp[
            (d_interp.step_index == 7) & (d_interp.step_index_counter == 4)
        ]
        self.assertLess(first_step.voltage.diff().max(), 0.001)
        self.assertLess(second_step.voltage.diff().max(), 0.001)

    # based on RCRT.test_get_diagnostic_summary
    def test_summarize_diagnostic(self):
        diag_summary = self.datapath_diag.summarize_diagnostic(self.diagnostic_available)
        self.assertEqual(diag_summary["paused"].max(), 0)
        self.assertEqual(diag_summary["CV_time"][0], np.float32(125502.578125))
        self.assertEqual(diag_summary["CV_current"][0], np.float32(2.3499656))
        self.assertEqual(diag_summary["CV_capacity"][0], np.float32(84.70442))

    # based on RCRT.test_determine_paused
    def test_paused_intervals(self):
        paused = self.datapath_paused.paused_intervals
        self.assertEqual(paused.max(), 7201.0)

        not_paused = self.datapath_diag.paused_intervals
        self.assertEqual(not_paused.max(), 0.0)

    # based on RCRT.test_get_diagnostic
    # though it is based on maccor files this test is designed to
    # check structuring of diagnostic cycles
    @unittest.skipUnless(BIG_FILE_TESTS, SKIP_MSG)
    def test_get_diagnostic(self):
        maccor_file_w_parameters_s3 = {
            "bucket": "beep-sync-test-stage",
            "key": "big_file_tests/PreDiag_000287_000128.092"
        }

        maccor_file_w_parameters = os.path.join(
            TEST_FILE_DIR, "PreDiag_000287_000128.092"
        )

        download_s3_object(bucket=maccor_file_w_parameters_s3["bucket"],
                           key=maccor_file_w_parameters_s3["key"],
                           destination_path=maccor_file_w_parameters)

        md = MaccorDatapath.from_file(maccor_file_w_parameters)

        (
            v_range,
            resolution,
            nominal_capacity,
            full_fast_charge,
            diagnostic_available,
        ) = md.determine_structuring_parameters()

        self.assertEqual(nominal_capacity, 4.84)
        # self.assertEqual(v_range, [2.7, 4.2]) # This is an older assertion, value changed when
        # different cell types were added

        self.assertEqual(v_range, [2.5, 4.2])
        self.assertEqual(
            diagnostic_available["cycle_type"],
            ["reset", "hppc", "rpt_0.2C", "rpt_1C", "rpt_2C"],
        )
        diag_summary = md.summarize_diagnostic(diagnostic_available)

        reg_summary = md.summarize_cycles(diagnostic_available)
        self.assertEqual(len(reg_summary.cycle_index.tolist()), 230)
        self.assertEqual(reg_summary.cycle_index.tolist()[:10],
                         [0, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        # Check data types are being set correctly for diagnostic summary
        diag_dyptes = diag_summary.dtypes.tolist()
        diag_columns = diag_summary.columns.tolist()
        diag_dyptes = [str(dtyp) for dtyp in diag_dyptes]
        for indx, col in enumerate(diag_columns):
            self.assertEqual(
                diag_dyptes[indx], STRUCTURE_DTYPES["diagnostic_summary"][col]
            )

        self.assertEqual(
            diag_summary.cycle_index.tolist(),
            [1, 2, 3, 4, 5, 36, 37, 38, 39, 40, 141, 142, 143, 144, 145, 246,
             247],
        )
        self.assertEqual(
            diag_summary.cycle_type.tolist(),
            [
                "reset",
                "hppc",
                "rpt_0.2C",
                "rpt_1C",
                "rpt_2C",
                "reset",
                "hppc",
                "rpt_0.2C",
                "rpt_1C",
                "rpt_2C",
                "reset",
                "hppc",
                "rpt_0.2C",
                "rpt_1C",
                "rpt_2C",
                "reset",
                "hppc",
            ],
        )
        self.assertEqual(diag_summary.paused.max(), 0)
        diag_interpolated = md.interpolate_diagnostic_cycles(
            diagnostic_available, resolution=1000
        )

        # Check data types are being set correctly for interpolated data
        diag_dyptes = diag_interpolated.dtypes.tolist()
        diag_columns = diag_interpolated.columns.tolist()
        diag_dyptes = [str(dtyp) for dtyp in diag_dyptes]
        for indx, col in enumerate(diag_columns):
            self.assertEqual(
                diag_dyptes[indx],
                STRUCTURE_DTYPES["diagnostic_interpolated"][col]
            )

        # Provide visual inspection to ensure that diagnostic interpolation is being done correctly
        diag_cycle = diag_interpolated[
            (diag_interpolated.cycle_type == "rpt_0.2C")
            & (diag_interpolated.step_type == 1)
            ]
        self.assertEqual(diag_cycle.cycle_index.unique().tolist(), [3, 38, 143])
        plt.figure()
        plt.plot(diag_cycle.discharge_capacity, diag_cycle.voltage)
        plt.savefig(
            os.path.join(TEST_FILE_DIR, "discharge_capacity_interpolation.png"))
        plt.figure()
        plt.plot(diag_cycle.voltage, diag_cycle.discharge_dQdV)
        plt.savefig(
            os.path.join(TEST_FILE_DIR, "discharge_dQdV_interpolation.png"))

        self.assertEqual(len(diag_cycle.index), 3000)

        hppcs = diag_interpolated[
            (diag_interpolated.cycle_type == "hppc")
            & pd.isnull(diag_interpolated.current)
            ]
        self.assertEqual(len(hppcs), 0)

        hppc_dischg1 = diag_interpolated[
            (diag_interpolated.cycle_index == 37)
            & (diag_interpolated.step_type == 2)
            & (diag_interpolated.step_index_counter == 3)
            & ~pd.isnull(diag_interpolated.current)
            ]

        plt.figure()
        plt.plot(hppc_dischg1.test_time, hppc_dischg1.voltage)
        plt.savefig(os.path.join(TEST_FILE_DIR, "hppc_discharge_pulse_1.png"))
        self.assertEqual(len(hppc_dischg1), 176)

        hppc_dischg2 = diag_interpolated[
            (diag_interpolated.cycle_index == 37)
            & (diag_interpolated.step_type == 6)
            # & (diag_interpolated.step_index_counter == 3)
            & ~pd.isnull(diag_interpolated.current)
            ]
        print(hppc_dischg2.step_type.unique())
        self.assertAlmostEqual(hppc_dischg2.voltage.min(), hppc_dischg2.voltage.max(), 3)
        plt.figure()
        plt.plot(hppc_dischg2.test_time, hppc_dischg2.current)
        plt.savefig(os.path.join(TEST_FILE_DIR, "hppc_cv.png"))
        self.assertEqual(len(hppc_dischg2), 1000)

        # processed_cycler_run = cycler_run.to_processed_cycler_run()
        md.autostructure()
        self.assertNotIn(
            diag_summary.cycle_index.tolist(),
            md.structured_data.cycle_index.unique(),
        )
        self.assertEqual(
            reg_summary.cycle_index.tolist(),
            md.structured_summary.cycle_index.tolist(),
        )

        processed_cycler_run_loc = os.path.join(
            TEST_FILE_DIR, "processed_diagnostic.json"
        )
        # Dump to the structured file and check the file size
        # File size had to be incteased as datapath dump includes ALL data now
        dumpfn(md, processed_cycler_run_loc)
        proc_size = os.path.getsize(processed_cycler_run_loc)
        self.assertLess(proc_size, 260000000)

        # Reload the structured file and check for errors
        test = loadfn(processed_cycler_run_loc)
        self.assertIsInstance(test.diagnostic_summary, pd.DataFrame)
        diag_dyptes = test.diagnostic_summary.dtypes.tolist()
        diag_columns = test.diagnostic_summary.columns.tolist()
        diag_dyptes = [str(dtyp) for dtyp in diag_dyptes]
        for indx, col in enumerate(diag_columns):
            self.assertEqual(
                diag_dyptes[indx], STRUCTURE_DTYPES["diagnostic_summary"][col]
            )

        diag_dyptes = test.diagnostic_data.dtypes.tolist()
        diag_columns = test.diagnostic_data.columns.tolist()
        diag_dyptes = [str(dtyp) for dtyp in diag_dyptes]
        for indx, col in enumerate(diag_columns):
            self.assertEqual(
                diag_dyptes[indx],
                STRUCTURE_DTYPES["diagnostic_interpolated"][col]
            )

        self.assertEqual(test.structured_summary.cycle_index.tolist()[:10],
                         [0, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        plt.figure()

        single_charge = test.structured_data[
            (test.structured_data.step_type == "charge")
            & (test.structured_data.cycle_index == 25)
            ]
        self.assertEqual(len(single_charge.index), 1000)
        plt.plot(single_charge.charge_capacity, single_charge.voltage)
        plt.savefig(
            os.path.join(
                TEST_FILE_DIR, "charge_capacity_interpolation_regular_cycle.png"
            )
        )

        os.remove(processed_cycler_run_loc)

    # based on PCRT.test_from_raw_cycler_run_parameters
    def test_structure(self):
        self.datapath_small_params.structure()
        self.assertEqual(self.datapath_small_params.metadata.barcode, "0001BC")
        self.assertEqual(self.datapath_small_params.metadata.protocol, "PredictionDiagnostics_000109.000")
        self.assertEqual(self.datapath_small_params.metadata.channel_id, 10)
        self.assertTrue(self.datapath_small_params.is_structured)

    def test_autostructure(self):
        self.datapath_diag_normal.autostructure()
        self.assertEqual(3448, len(self.datapath_diag_normal.structured_summary))
        self.assertEqual(4957, self.datapath_diag_normal.structured_summary.paused.iloc[0])

        self.assertEqual("reset", self.datapath_diag_normal.diagnostic_summary.cycle_type.iloc[0])
        self.assertEqual(4.711819, np.round(self.datapath_diag_normal.diagnostic_summary.discharge_capacity.iloc[0], 6))

    # based on PCRT.test_get_cycle_life
    def test_get_cycle_life(self):
        datapath = self.BEEPDatapathChildTest.from_json_file(self.cycle_run_file)
        self.assertEqual(datapath.get_cycle_life(30, 0.99), 82)
        self.assertEqual(datapath.get_cycle_life(40, 0.0), 189)

    # based on PCRT.test_data_types_old_processed
    def test_data_types_old_processed(self):
        datapath = self.BEEPDatapathChildTest.from_json_file(self.cycle_run_file)

        all_summary = datapath.structured_summary
        reg_dyptes = all_summary.dtypes.tolist()
        reg_columns = all_summary.columns.tolist()
        reg_dyptes = [str(dtyp) for dtyp in reg_dyptes]
        for indx, col in enumerate(reg_columns):
            self.assertEqual(reg_dyptes[indx], STRUCTURE_DTYPES["summary"][col])

        all_interpolated = datapath.structured_data
        cycles_interpolated_dyptes = all_interpolated.dtypes.tolist()
        cycles_interpolated_columns = all_interpolated.columns.tolist()
        cycles_interpolated_dyptes = [str(dtyp) for dtyp in cycles_interpolated_dyptes]
        for indx, col in enumerate(cycles_interpolated_columns):
            self.assertEqual(
                cycles_interpolated_dyptes[indx],
                STRUCTURE_DTYPES["cycles_interpolated"][col],
            )

    # based on PCRT.test_cycles_to_reach_set_capacities
    def test_capacities_to_cycles(self):
        datapath = self.BEEPDatapathChildTest.from_json_file(self.cycle_run_file)
        cycles = datapath.capacities_to_cycles()
        self.assertGreaterEqual(cycles.iloc[0, 0], 100)

    # based on PCRT.test_capacities_at_set_cycles
    def test_cycles_to_capacities(self):
        datapath = self.BEEPDatapathChildTest.from_json_file(self.cycle_run_file)
        capacities = datapath.cycles_to_capacities()
        self.assertLessEqual(capacities.iloc[0, 0], 1.1)

    def test_semiunique_id(self):
        dp = copy.deepcopy(self.datapath_small_params)

        # Ensure repeated calls produce the same id
        self.assertEqual(dp.semiunique_id, dp.semiunique_id)

        if dp.is_structured:
            dp.unstructure()
        h_unstructured = dp.semiunique_id

        dp.structure()

        # Ensure repeated calls return the same hash (deterministic somewhat)
        h_structured = dp.semiunique_id
        self.assertNotEqual(h_unstructured, h_structured)

        # Changing paths must change the semiunique id
        # As differently datapaths with identical raw but different structuring
        # data saved under different filenames must appear as different
        with ScratchDir("."):
            dp.to_json_file("dp_test_hash.json", omit_raw=True)
            dp = self.BEEPDatapathChildTest.from_json_file("dp_test_hash.json")

            self.assertNotEqual(dp.semiunique_id, h_structured)


class TestBaseEIS(unittest.TestCase):
    class EISChildGood(EIS):
        def from_file(self):
            print("success EISChildGood")

    def setUp(self) -> None:
        self.raw_data = pd.DataFrame({"a": [1,2,3]})
        self.metadata = pd.DataFrame({"example": ["metadata"]})

    def test_BEEPDatapathWithEIS(self):

        # Must implement load_eis and from_file
        class BEEPDatapathWithEISChildTestGood(BEEPDatapathWithEIS):
            def from_file(self, path):
                print(f"{path}")

            def load_eis(self, *arg, **kwargs):
                print("success BEEPDatapathWithEISChildTestGood")

        class BEEPDatapathWithEISChildTestBad(BEEPDatapathWithEIS):
            def extra_method(self):
                pass

        example_metadata = {"example": "metadata"}

        BEEPDatapathWithEISChildTestGood(raw_data=self.raw_data, metadata=example_metadata)

        with self.assertRaises(TypeError):
            BEEPDatapathWithEISChildTestBad(raw_data=self.raw_data, metadata=example_metadata)

    def test_EIS(self):
        # must implement a from_file method
        class EISChildBad(EIS):
            def extra_method(self):
                pass

        eis = self.EISChildGood(data=self.raw_data, metadata=self.metadata)
        self.assertTrue(hasattr(eis, "from_dict"))
        self.assertTrue(hasattr(eis, "as_dict"))

        with self.assertRaises(TypeError):
            EISChildBad()

    def test_EIS_serialization(self):
        eis = self.EISChildGood(data=self.raw_data, metadata=self.metadata)
        d = eis.as_dict()

        eis_test = self.EISChildGood.from_dict(d)
        self.assertTrue(eis.data.equals(eis_test.data))
        self.assertTrue(eis.metadata.equals(eis_test.metadata))


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
from scipy.stats import skew, kurtosis
from scipy.interpolate import interp1d

from beep import PROTOCOL_PARAMETERS_DIR
from beep.features import featurizer_helpers
from beep.features.base import BEEPFeaturizer, BEEPFeaturizationError


class HPPCResistanceVoltageFeatures(BEEPFeaturizer):
    DEFAULT_HYPERPARAMETERS = {
        "test_time_filter_sec": 1000000,
        "cycle_index_filter": 6,
        "diag_pos": 1,
        "soc_window": 8,
        "parameters_path": PROTOCOL_PARAMETERS_DIR
    }

    def validate(self):
        val, msg = featurizer_helpers.check_diagnostic_validation(self.datapath)
        if val:
            conditions = []
            conditions.append(
                any(
                    [
                        "hppc" in x
                        for x in
                        self.datapath.diagnostic_summary.cycle_type.unique()
                    ]
                )
            )
            if all(conditions):
                return True, None
            else:
                return False, "HPPC conditions not met for this cycler run"
        else:
            return val, msg

    def create_features(self):
        # Filter out low cycle numbers at the end of the test, corresponding to the "final" diagnostic
        self.datapath.diagnostic_data = self.datapath.diagnostic_data[
            ~((self.datapath.diagnostic_data.test_time > self.hyperparameters[
                'test_time_filter_sec']) &
              (self.datapath.diagnostic_data.cycle_index < self.hyperparameters[
                  'cycle_index_filter']))
        ]
        self.datapath.diagnostic_data = self.datapath.diagnostic_data.groupby(
            ["cycle_index", "step_index", "step_index_counter"]
        ).filter(lambda x: ~x["test_time"].isnull().all())

        # diffusion features
        diffusion_features = featurizer_helpers.get_diffusion_features(
            self.datapath, self.hyperparameters["diag_pos"]
        )

        hppc_r = pd.DataFrame()
        # the 9 by 6 dataframe
        df_dr = featurizer_helpers.get_dr_df(
            self.datapath, self.hyperparameters["diag_pos"]
        )
        # transform this dataframe to be 1 by 54
        columns = df_dr.columns
        for column in columns:
            for r in range(len(df_dr[column])):
                name = column + str(r)
                hppc_r[name] = [df_dr[column][r]]

        # the variance of ocv features
        hppc_ocv = featurizer_helpers.get_hppc_ocv(
            self.datapath,
            self.hyperparameters["diag_pos"],
            parameters_path=self.hyperparameters["parameters_path"]
        )

        # the v_diff features
        v_diff = featurizer_helpers.get_v_diff(
            self.datapath,
            self.hyperparameters["diag_pos"],
            self.hyperparameters["soc_window"],
            self.hyperparameters["parameters_path"]
        )

        # merge everything together as a final result dataframe
        self.features = pd.concat(
            [hppc_r, hppc_ocv, v_diff, diffusion_features], axis=1)


class CycleSummaryStats(BEEPFeaturizer):
    DEFAULT_HYPERPARAMETERS = {
        "cycle_comp_num": [10, 100],
        "statistics": ["var", "min", "mean", "skew", "kurtosis", "abs",
                       "square"]
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """

        # TODO: not sure this is necessary
        # Check for data in each of the selected cycles
        index_1, index_2 = self.hyperparameters['cycle_comp_num']
        cycle_1 = self.datapath.structured_data[
            self.datapath.structured_data.cycle_index == index_1]
        cycle_2 = self.datapath.structured_data[
            self.datapath.structured_data.cycle_index == index_2]
        if len(cycle_1) == 0 or len(cycle_2) == 0:
            return False, "Length of one or more comparison cycles is zero"

        # TODO: check whether this is good
        # Check for relevant data
        required_columns = [
            'charge_capacity',
            'discharge_capacity',
            'charge_energy',
            'discharge_energy',
        ]
        pcycler_run_columns = self.datapath.structured_data.columns
        if not all(
                [column in pcycler_run_columns for column in required_columns]):
            return False, f"Required column not present in all structured data " \
                          f"(must have all of: {required_columns})"

        return True, None

    def create_features(self):
        """
        Generate features listed in early prediction manuscript using both diagnostic and regular cycles

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun)
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            X (pd.DataFrame): Dataframe containing the feature
        """

        # TODO: extend this dataframe and uncomment energy features when
        #   structuring is refactored
        X = pd.DataFrame(np.zeros((1, 28)))

        reg_cycle_comp_num = self.hyperparameters.get("cycle_comp_num")
        cycle_comp_1 = self.datapath.structured_data[
            self.datapath.structured_data.cycle_index == reg_cycle_comp_num[1]
            ]
        cycle_comp_0 = self.datapath.structured_data[
            self.datapath.structured_data.cycle_index == reg_cycle_comp_num[0]
            ]
        Qc100_1 = cycle_comp_1[
            cycle_comp_1.step_type == "charge"].charge_capacity
        Qc10_1 = cycle_comp_0[
            cycle_comp_0.step_type == "charge"].charge_capacity
        QcDiff = Qc100_1.values - Qc10_1.values
        QcDiff = QcDiff[~np.isnan(QcDiff)]

        X.loc[0, 0:6] = self.get_summary_statistics(QcDiff)

        Qd100_1 = cycle_comp_1[
            cycle_comp_1.step_type == "discharge"].discharge_capacity
        Qd10_1 = cycle_comp_0[
            cycle_comp_0.step_type == "discharge"].discharge_capacity
        QdDiff = Qd100_1.values - Qd10_1.values
        QdDiff = QdDiff[~np.isnan(QdDiff)]

        X.loc[0, 7:13] = self.get_summary_statistics(QdDiff)

        # # Charging Energy features
        Ec100_1 = cycle_comp_1[cycle_comp_1.step_type == "charge"].charge_energy
        Ec10_1 = cycle_comp_0[cycle_comp_0.step_type == "charge"].charge_energy
        EcDiff = Ec100_1.values - Ec10_1.values
        EcDiff = EcDiff[~np.isnan(EcDiff)]

        X.loc[0, 14:20] = self.get_summary_statistics(EcDiff)

        # # Discharging Energy features
        Ed100_1 = cycle_comp_1[
            cycle_comp_1.step_type == "charge"].discharge_energy
        Ed10_1 = cycle_comp_0[
            cycle_comp_0.step_type == "charge"].discharge_energy
        EdDiff = Ed100_1.values - Ed10_1.values
        EdDiff = EdDiff[~np.isnan(EdDiff)]

        X.loc[0, 21:27] = self.get_summary_statistics(EdDiff)

        quantities = [
            "charging_capacity",
            "discharging_capacity",
            "charging_energy",
            "discharging_energy",
        ]

        X.columns = [y + "_" + x for x in quantities for y in
                     self.hyperparameters["statistics"]]

        self.features = X

    def get_summary_statistics(self, array):
        """
        Static method for getting values corresponding
        to standard 7 operations that many beep features
        use, i.e. log of absolute value of each of
        variance, min, mean, skew, kurtosis, the sum of
        the absolute values and the sum of squares

        Args:
            array (list, np.ndarray): array of values to get
                standard operation values for, e.g. cycle
                discharging capacity, QcDiff, etc.

        Returns:
            [float]: list of features

        """

        stats_names = self.hyperparameters["statistics"]
        supported_stats = self.DEFAULT_HYPERPARAMETERS["statistics"]

        if any(s not in supported_stats for s in stats_names):
            raise ValueError(
                f"Unsupported statistics in {stats_names}: supported statistics are {supported_stats}")

        stats = []

        if "var" in stats_names:
            stats.append(np.log10(np.absolute(np.var(array))))
        if "min" in stats_names:
            stats.append(np.log10(np.absolute(min(array))))
        if "mean" in stats_names:
            stats.append(np.log10(np.absolute(np.mean(array))))
        if "skew" in stats_names:
            stats.append(np.log10(np.absolute(skew(array))))
        if "kurtosis" in stats_names:
            stats.append(np.log10(
                np.absolute(kurtosis(array, fisher=False, bias=False))))
        if "abs" in stats_names:
            stats.append(np.log10(np.sum(np.absolute(array))))
        if "square" in stats_names:
            stats.append(np.log10(np.sum(np.square(array))))

        return np.asarray(stats)


class DiagnosticSummaryStats(CycleSummaryStats):
    """
    Object corresponding to summary statistics from a diagnostic cycle of
    specific type. Includes constructors to create the features, object names
    and metadata attributes in the object.  Inherits from RegularCycleSummaryStats
    to reuse standard feature generation

    name (str): predictor object name.
    X (pandas.DataFrame): features in DataFrame format.
    metadata (dict): information about the conditions, data
        and code used to produce features
    """
    DEFAULT_HYPERPARAMETERS = {
        "test_time_filter_sec": 1000000,
        "cycle_index_filter": 6,
        "diagnostic_cycle_type": 'rpt_0.2C',
        "diag_pos_list": [0, 1],
        "statistics": ["var", "min", "mean", "skew", "kurtosis", "abs",
                       "square"],
        "parameters_path": PROTOCOL_PARAMETERS_DIR
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations

        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        val, msg = featurizer_helpers.check_diagnostic_validation(self.datapath)
        if val:
            df = self.datapath.diagnostic_summary
            df = df[
                df.cycle_type == self.hyperparameters["diagnostic_cycle_type"]]
            if df.cycle_index.nunique() >= max(
                    self.hyperparameters["diag_pos_list"]) + 1:
                return True, None
            else:
                return False, "Diagnostic cycles insufficient for featurization"
        else:
            return val, msg

    def get_summary_diff(
            self,
            pos=None,
            cycle_types=("rpt_0.2C", "rpt_1C", "rpt_2C"),
            metrics=(
                    "discharge_capacity", "discharge_energy", "charge_capacity",
                    "charge_energy")
    ):
        """
        Helper function to calculate difference between summary values in the diagnostic cycles

                Args:
                    processed_cycler_run (beep.structure.ProcessedCyclerRun)
                    pos (list): position of the diagnostics to use in the calculation
                    cycle_types (list): calculate difference for these diagnostic types
                    metrics (str): Calculate difference for these metrics

                Returns:
                    values (list): List of difference values to insert into the dataframe
                    names (list): List of column headers to use in the creation of the dataframe
                """
        pos = self.hyperparameters["diag_pos_list"] if not pos else pos

        values = []
        names = []
        for cycle_type in cycle_types:
            diag_type_summary = self.datapath.diagnostic_summary[
                self.datapath.diagnostic_summary.cycle_type == cycle_type]
            for metric in metrics:
                diff = (diag_type_summary.iloc[pos[1]][metric] -
                        diag_type_summary.iloc[pos[0]][metric]) \
                       / diag_type_summary.iloc[pos[0]][metric]
                values.append(diff)
                names.append("diag_sum_diff_" + str(pos[0]) + "_" + str(
                    pos[1]) + "_" + cycle_type + metric)
        return values, names

    def create_features(self):
        """
        Generate features listed in early prediction manuscript using both diagnostic and regular cycles

        Args:
            self.datapathn (beep.structure.ProcessedCyclerRun)
            self.hyperparameters (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            X (pd.DataFrame): Dataframe containing the feature
        """
        # Filter out "final" diagnostic cycles that have been appended to the end of the file with the wrong
        # cycle number(test time is monotonic)
        self.datapath.diagnostic_data = self.datapath.diagnostic_data[
            ~((self.datapath.diagnostic_data.test_time > self.hyperparameters[
                'test_time_filter_sec']) &
              (self.datapath.diagnostic_data.cycle_index < self.hyperparameters[
                  'cycle_index_filter']))
        ]
        self.datapath.diagnostic_data = self.datapath.diagnostic_data.groupby(
            ["cycle_index", "step_index", "step_index_counter"]
        ).filter(lambda x: ~x["test_time"].isnull().all())

        diag_intrp = self.datapath.diagnostic_data

        X = pd.DataFrame(np.zeros((1, 54)))

        # Calculate the cycles and the steps for the selected diagnostics
        cycles = diag_intrp.cycle_index[diag_intrp.cycle_type ==
                                        self.hyperparameters[
                                            "diagnostic_cycle_type"]].unique()
        step_dict_0 = featurizer_helpers.get_step_index(
            self.datapath,
            cycle_type=self.hyperparameters["diagnostic_cycle_type"],
            diag_pos=self.hyperparameters["diag_pos_list"][0],
            parameters_path=self.hyperparameters["parameters_path"]
        )
        step_dict_1 = featurizer_helpers.get_step_index(
            self.datapath,
            cycle_type=self.hyperparameters["diagnostic_cycle_type"],
            diag_pos=self.hyperparameters["diag_pos_list"][1],
            parameters_path=self.hyperparameters["parameters_path"]
        )

        # Create masks for each position in the data
        mask_pos_0_charge = ((diag_intrp.cycle_index == cycles[
            self.hyperparameters["diag_pos_list"][0]]) &
                             (diag_intrp.step_index == step_dict_0[
                                 self.hyperparameters[
                                     "diagnostic_cycle_type"] + '_charge']))
        mask_pos_1_charge = ((diag_intrp.cycle_index == cycles[
            self.hyperparameters["diag_pos_list"][1]]) &
                             (diag_intrp.step_index == step_dict_1[
                                 self.hyperparameters[
                                     "diagnostic_cycle_type"] + '_charge']))
        mask_pos_0_discharge = ((diag_intrp.cycle_index == cycles[
            self.hyperparameters["diag_pos_list"][0]]) &
                                (diag_intrp.step_index ==
                                 step_dict_0[self.hyperparameters[
                                                 "diagnostic_cycle_type"] + '_discharge']))
        mask_pos_1_discharge = ((diag_intrp.cycle_index == cycles[
            self.hyperparameters["diag_pos_list"][1]]) &
                                (diag_intrp.step_index ==
                                 step_dict_1[self.hyperparameters[
                                                 "diagnostic_cycle_type"] + '_discharge']))

        # Charging Capacity features
        Qc_1 = diag_intrp.charge_capacity[mask_pos_1_charge]
        Qc_0 = diag_intrp.charge_capacity[mask_pos_0_charge]
        QcDiff = Qc_1.values - Qc_0.values
        QcDiff = QcDiff[~np.isnan(QcDiff)]

        X.loc[0, 0:6] = self.get_summary_statistics(QcDiff)

        # Discharging Capacity features
        Qd_1 = diag_intrp.discharge_capacity[mask_pos_1_discharge]
        Qd_0 = diag_intrp.discharge_capacity[mask_pos_0_discharge]
        QdDiff = Qd_1.values - Qd_0.values
        QdDiff = QdDiff[~np.isnan(QdDiff)]

        X.loc[0, 7:13] = self.get_summary_statistics(QdDiff)

        # Charging Energy features
        Ec_1 = diag_intrp.charge_energy[mask_pos_1_charge]
        Ec_0 = diag_intrp.charge_energy[mask_pos_0_charge]
        EcDiff = Ec_1.values - Ec_0.values
        EcDiff = EcDiff[~np.isnan(EcDiff)]

        X.loc[0, 14:20] = self.get_summary_statistics(EcDiff)

        # Discharging Energy features
        Ed_1 = diag_intrp.discharge_energy[mask_pos_1_discharge]
        Ed_0 = diag_intrp.discharge_energy[mask_pos_0_discharge]
        EdDiff = Ed_1.values - Ed_0.values
        EdDiff = EdDiff[~np.isnan(EdDiff)]

        X.loc[0, 21:27] = self.get_summary_statistics(EdDiff)

        # Charging dQdV features
        dQdVc_1 = diag_intrp.charge_dQdV[mask_pos_1_charge]
        dQdVc_0 = diag_intrp.charge_dQdV[mask_pos_0_charge]
        dQdVcDiff = dQdVc_1.values - dQdVc_0.values
        dQdVcDiff = dQdVcDiff[~np.isnan(dQdVcDiff)]

        X.loc[0, 28:34] = self.get_summary_statistics(dQdVcDiff)

        # Discharging Capacity features
        dQdVd_1 = diag_intrp.discharge_dQdV[mask_pos_1_discharge]
        dQdVd_0 = diag_intrp.discharge_dQdV[mask_pos_0_discharge]
        dQdVdDiff = dQdVd_1.values - dQdVd_0.values
        dQdVdDiff = dQdVdDiff[~np.isnan(dQdVdDiff)]

        X.loc[0, 35:41] = self.get_summary_statistics(dQdVdDiff)

        X.loc[0, 42:53], names = self.get_summary_diff(
            self.hyperparameters["diag_pos_list"]
        )

        quantities = [
            "charging_capacity",
            "discharging_capacity",
            "charging_energy",
            "discharging_energy",
            "charging_dQdV",
            "discharging_dQdV",
        ]

        X.columns = [y + "_" + x for x in quantities for y in
                     self.hyperparameters["statistics"]] + names
        self.features = X


class DeltaQFastCharge(BEEPFeaturizer):
    """
    Object corresponding to feature object. Includes constructors
    to create the features, object names and metadata attributes in the
    object
        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features
    """
    DEFAULT_HYPERPARAMETERS = {
        "init_pred_cycle": 10,
        "mid_pred_cycle": 91,
        "final_pred_cycle": 100,
        "n_nominal_cycles": 40
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            self.datapath (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """

        if not self.datapath.structured_summary.index.max() > self.hyperparameters["final_pred_cycle"]:
            return False, "Structured summary index max is less than final pred cycle"
        elif not self.datapath.structured_summary.index.min() <= self.hyperparameters["init_pred_cycle"]:
            return False, "Structured summary index min is more than initial pred cycle"
        elif "cycle_index" not in self.datapath.structured_summary.columns:
            return False, "Structured summary missing critical data: 'cycle_index'"
        elif "cycle_index" not in self.datapath.structured_data.columns:
            return False, "Structured data missing critical data: 'cycle_index'"
        elif not self.hyperparameters["mid_pred_cycle"] > 10:
            return False, "Middle pred. cycle less than threshold value of 10"
        elif not self.hyperparameters["final_pred_cycle"] > self.hyperparameters["mid_pred_cycle"]:
            return False, "Final pred cycle less than middle pred cycle"
        else:
            return True, None

    def create_features(self):
        """
        Generate features listed in early prediction manuscript, primarily related to the
        so called delta Q feature
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            self.hyperparameters (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: features indicative of degradation, derived from the input data
        """
        i_final = self.hyperparameters[
                      "final_pred_cycle"] - 1  # python indexing
        i_mid = self.hyperparameters["mid_pred_cycle"] - 1

        summary = self.datapath.structured_summary
        self.hyperparameters[
            "n_nominal_cycles"
        ] = 40  # For nominal capacity, use median discharge capacity of first n cycles

        if "step_type" in self.datapath.structured_data.columns:
            interpolated_df = self.datapath.structured_data[
                self.datapath.structured_data.step_type == "discharge"
                ]
        else:
            interpolated_df = self.datapath.structured_data
        X = pd.DataFrame(np.zeros((1, 20)))
        labels = []
        # Discharge capacity, cycle 2 = Q(n=2)
        X[0] = summary.discharge_capacity.iloc[1]
        labels.append("discharge_capacity_cycle_2")

        # Max discharge capacity - discharge capacity, cycle 2 = max_n(Q(n)) - Q(n=2)
        X[1] = max(
            summary.discharge_capacity.iloc[np.arange(i_final + 1)]
            - summary.discharge_capacity.iloc[1]
        )
        labels.append("max_discharge_capacity_difference")

        # Discharge capacity, cycle 100 = Q(n=100)
        X[2] = summary.discharge_capacity.iloc[i_final]
        labels.append("discharge_capacity_cycle_100")

        # Feature representing time-temperature integral over cycles 2 to 100
        X[3] = np.nansum(
            summary.time_temperature_integrated.iloc[np.arange(i_final + 1)])
        labels.append("integrated_time_temperature_cycles_1:100")

        # Mean of charge times of first 5 cycles
        X[4] = np.nanmean(summary.charge_duration.iloc[1:6])
        labels.append("charge_time_cycles_1:5")

        # Descriptors based on capacity loss between cycles 10 and 100.
        Qd_final = interpolated_df.discharge_capacity[
            interpolated_df.cycle_index == i_final
            ]
        Qd_10 = interpolated_df.discharge_capacity[
            interpolated_df.cycle_index == 9]

        Qd_diff = Qd_final.values - Qd_10.values

        # If DeltaQ(V) is not an empty array, compute summary stats, else initialize with np.nan
        # Cells discharged rapidly over a narrow voltage window run into have no interpolated discharge steps
        if len(Qd_diff):
            X[5] = np.log10(np.abs(np.nanmin(Qd_diff)))  # Minimum
            X[6] = np.log10(np.abs(np.nanmean(Qd_diff)))  # Mean
            X[7] = np.log10(np.abs(np.nanvar(Qd_diff)))  # Variance
            X[8] = np.log10(np.abs(skew(Qd_diff)))  # Skewness
            X[9] = np.log10(np.abs(kurtosis(Qd_diff)))  # Kurtosis
            X[10] = np.log10(np.abs(Qd_diff[0]))  # First difference
        else:
            X[5:11] = np.nan

        labels.append("abs_min_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_mean_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_variance_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_skew_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_kurtosis_discharge_capacity_difference_cycles_2:100")
        labels.append("abs_first_discharge_capacity_difference_cycles_2:100")

        X[11] = np.max(summary.temperature_maximum.iloc[
                           list(range(1, i_final + 1))])  # Max T
        labels.append("max_temperature_cycles_1:100")

        X[12] = np.min(summary.temperature_minimum.iloc[
                           list(range(1, i_final + 1))])  # Min T
        labels.append("min_temperature_cycles_1:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 2 to 100

        X[13], X[14] = np.polyfit(
            list(range(1, i_final + 1)),
            summary.discharge_capacity.iloc[list(range(1, i_final + 1))],
            1,
        )

        labels.append("slope_discharge_capacity_cycle_number_2:100")
        labels.append("intercept_discharge_capacity_cycle_number_2:100")

        # Slope and intercept of linear fit to discharge capacity as a fn of cycle #, cycles 91 to 100
        X[15], X[16] = np.polyfit(
            list(range(i_mid, i_final + 1)),
            summary.discharge_capacity.iloc[list(range(i_mid, i_final + 1))],
            1,
        )
        labels.append("slope_discharge_capacity_cycle_number_91:100")
        labels.append("intercept_discharge_capacity_cycle_number_91:100")

        IR_trend = summary.dc_internal_resistance.iloc[
            list(range(1, i_final + 1))]
        if any(v == 0 for v in IR_trend):
            IR_trend[IR_trend == 0] = np.nan

        # Internal resistance minimum
        X[17] = np.nanmin(IR_trend)
        labels.append("min_internal_resistance_cycles_2:100")

        # Internal resistance at cycle 2
        X[18] = summary.dc_internal_resistance.iloc[1]
        labels.append("internal_resistance_cycle_2")

        # Internal resistance at cycle 100 - cycle 2
        X[19] = (
                summary.dc_internal_resistance.iloc[i_final] -
                summary.dc_internal_resistance.iloc[1]
        )
        labels.append("internal_resistance_difference_cycles_2:100")

        # Nominal capacity
        end = self.hyperparameters["n_nominal_cycles"]
        X[20] = np.median(summary.discharge_capacity.iloc[0: end])
        labels.append("nominal_capacity_by_median")

        X.columns = labels
        self.features = X


class TrajectoryFastCharge(DeltaQFastCharge):
    """
    Object corresponding to cycle numbers at which the capacity drops below
     specific percentages of the initial capacity. Computed on the discharge
     portion of the regular fast charge cycles.

    """

    DEFAULT_HYPERPARAMETERS = {
        "thresh_max_cap": 0.98,
        "thresh_min_cap": 0.78,
        "interval_cap": 0.03
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        cap = self.datapath.structured_summary.discharge_capacity
        cap_ratio = cap.min() / cap.max()
        max_cap = self.hyperparameters["thresh_max_cap"]
        if not cap_ratio < max_cap:
            return False, f"thresh_max_cap hyperparameter exceeded: {cap_ratio} !< {max_cap}"
        else:
            return True, None

    def create_features(self):
        """
        Calculate the outcomes from the input data. In particular, the number of cycles
        where we expect to reach certain thresholds of capacity loss
        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: cycles at which capacity/energy degradation exceeds thresholds
        """
        y = self.datapath.capacities_to_cycles(
            self.hyperparameters["thresh_max_cap"],
            self.hyperparameters["thresh_min_cap"],
            self.hyperparameters["interval_cap"],
        )
        self.features = y


class DiagnosticProperties(BEEPFeaturizer):
    """
    This class stores fractional levels of degradation in discharge capacity and discharge energy
    relative to the first cycle at each diagnostic cycle, grouped by diagnostic cycle type.

        name (str): predictor object name.
        X (pandas.DataFrame): features in DataFrame format.
        metadata (dict): information about the conditions, data
            and code used to produce features

    Hyperparameters:
        parameters_dir (str): Full path to directory of parameters to analyse the
            diagnostic cycles
        quantities ([str]): Quantities to extract/get fractional metrics for
            diagnostic cycles
        cycle_type (str): Type of diagnostic cycle being used to measure the
            fractional metric
        metric (str): The metric being used for fractional capacity
        interpolation_axes (list): List of column names to use for
            x_axis interpolation (distance to threshold)
        threshold (float): Value for the fractional metric to be considered above
            or below threshold
        filter_kinks (float): If set, cutoff value for the second derivative of
            the fractional metric (cells with an abrupt change in degradation
            rate might have something else going on). Typical value might be 0.04
        extrapolate_threshold (bool): Should threshold crossing point be
            extrapolated for cells that have not yet reached the threshold
            (warning: this uses a linear extrapolation from the last two
            diagnostic cycles)
    """
    DEFAULT_HYPERPARAMETERS = {
        "parameters_dir": PROTOCOL_PARAMETERS_DIR,
        "quantities": ['discharge_energy', 'discharge_capacity'],
        "threshold": 0.8,
        "metric": "discharge_energy",
        "filter_kinks": None,
        "interpolation_axes": ["normalized_regular_throughput", "cycle_index"],
        "cycle_type": "rpt_1C",
        "extrapolate_threshold": True
    }

    def validate(self):
        """
        This function determines if the input data has the necessary attributes for
        creation of this feature class. It should test for all of the possible reasons
        that feature generation would fail for this particular input data.

        Args:
            processed_cycler_run (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
            gets featurized. These could be filters for column or row operations
        Returns:
            bool: True/False indication of ability to proceed with feature generation
        """
        return featurizer_helpers.check_diagnostic_validation(self.datapath)

    def create_features(self):
        """
        Generates diagnostic-property features from processed cycler run, including values for n*x method
        Args:
            self.datapath (beep.structure.ProcessedCyclerRun): data from cycler run
            params_dict (dict): dictionary of parameters governing how the ProcessedCyclerRun object
                gets featurized. These could be filters for column or row operations
            parameters_path (str): Root directory storing project parameter files.

        Returns:
            pd.DataFrame: with "cycle_index", "fractional_metric", "x", "n", "cycle_type" and "metric" columns, rows
            for each diagnostic cycle of the cell
        """

        parameters_path = self.hyperparameters["parameters_dir"]

        cycle_types = self.datapath.diagnostic_summary.cycle_type.unique()
        X = pd.DataFrame()
        for quantity in self.hyperparameters["quantities"]:
            for cycle_type in cycle_types:
                summary_diag_cycle_type = featurizer_helpers.get_fractional_quantity_remaining_nx(
                    self.datapath, quantity, cycle_type,
                    parameters_path=parameters_path
                )

                summary_diag_cycle_type.loc[:, "cycle_type"] = cycle_type
                summary_diag_cycle_type.loc[:, "metric"] = quantity
                X = X.append(summary_diag_cycle_type)

        X_condensed = self.get_threshold_targets(X)
        self.features = X_condensed

    def get_threshold_targets(self, df):
        """
        Apply a threshold via interpolation for determining various
        metrics (e.g., discharge energy) from diagnostic cycles.

        Args:
            df (pd.DataFrame): A dataframe of diagnostic cycle data
                for a single battery cycler run.

        Returns:
            (pd.DataFrame): Contains a vector for interpolated/intercept
                data for determining threshold.

        """
        cycle_type = self.hyperparameters["cycle_type"]
        metric = self.hyperparameters["metric"]
        interpolation_axes = self.hyperparameters["interpolation_axes"]
        threshold = self.hyperparameters["threshold"]
        filter_kinks = self.hyperparameters["filter_kinks"]
        extrapolate_threshold = self.hyperparameters["extrapolate_threshold"]

        if filter_kinks:
            if np.any(df['fractional_metric'].diff().diff() < filter_kinks):
                last_good_cycle = df[
                    df['fractional_metric'].diff().diff() < filter_kinks]['cycle_index'].min()
                df = df[df['cycle_index'] < last_good_cycle]

        x_axes = []
        for type in interpolation_axes:
            x_axes.append(df[type])
        y_interpolation_axis = df['fractional_metric']

        # Logic around how to deal with cells that have not crossed threshold
        if df['fractional_metric'].min() > threshold and \
                not extrapolate_threshold:
            BEEPFeaturizationError(
                "DiagnosticProperties data has not crossed threshold "
                "and extrapolation inaccurate"
            )
        elif df['fractional_metric'].min() > threshold and \
                extrapolate_threshold:
            fill_value = "extrapolate"
            bounds_error = False
            x_linspaces = []
            for x_axis in x_axes:
                y1 = y_interpolation_axis.iloc[-2]
                y2 = y_interpolation_axis.iloc[-1]
                x1 = x_axis.iloc[-2]
                x2 = x_axis.iloc[-1]
                x_thresh_extrap = (threshold - 0.1 - y1) * (x2 - x1) / (
                        y2 - y1) + x1
                x_linspaces.append(
                    np.linspace(x_axis.min(), x_thresh_extrap, num=1000)
                )
        else:
            fill_value = np.nan
            bounds_error = True
            x_linspaces = []
            for x_axis in x_axes:
                x_linspaces.append(
                    np.linspace(x_axis.min(), x_axis.max(), num=1000))

        f_axis = []
        for x_axis in x_axes:
            f_axis.append(
                interp1d(
                    x_axis,
                    y_interpolation_axis,
                    kind='linear',
                    bounds_error=bounds_error,
                    fill_value=fill_value
                )
            )

        x_to_threshold = []
        for indx, x_linspace in enumerate(x_linspaces):
            crossing_array = abs(f_axis[indx](x_linspace) - threshold)
            x_to_threshold.append(x_linspace[np.argmin(crossing_array)])

        if ~(x_to_threshold[0] > 0) or ~(x_to_threshold[1] > 0):
            raise BEEPFeaturizationError(
                "DiagnosticProperties data does not have a positive value "
                "to threshold"
            )

        if "normalized_regular_throughput" in interpolation_axes:
            real_throughput_to_threshold = x_to_threshold[
                                               interpolation_axes.index(
                                                   "normalized_regular_throughput")] * \
                                           df[
                                               'initial_regular_throughput'].values[
                                               0]
            x_to_threshold.append(real_throughput_to_threshold)
            interpolation_axes = interpolation_axes + ["real_regular_throughput"]

        threshold_dict = {
            'initial_regular_throughput':
                df['initial_regular_throughput'].values[0],
        }

        for indx, x_axis in enumerate(interpolation_axes):
            threshold_dict[
                cycle_type + metric + str(threshold) + '_' + x_axis] = [
                x_to_threshold[indx]]

        return pd.DataFrame(threshold_dict)



# Copyright [2020] [Toyota Research Institute]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests related to batch validation"""

import json
import os
import unittest
import tempfile
import pandas as pd
import numpy as np
from pathlib import Path
from monty.tempfile import ScratchDir
from beep.structure.validate import SimpleValidator
from beep import S3_CACHE, VALIDATION_SCHEMA_DIR
from beep.tests.constants import TEST_FILE_DIR


@unittest.skip("Needs to be integrated with new structuring")
class ValidationMaccorTest(unittest.TestCase):
    # To further develop as Maccor data / schema becomes available
    def setUp(self):
        pass

    def test_validation_maccor(self):
        path = "xTESLADIAG_000019_CH70.070"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator(
            schema_filename=os.path.join(
                VALIDATION_SCHEMA_DIR, "schema-maccor-2170.yaml"
            )
        )
        v.allow_unknown = True
        header = pd.read_csv(path, delimiter="\t", nrows=0)
        df = pd.read_csv(path, delimiter="\t", skiprows=1)
        df["State"] = df["State"].astype(str)
        df["current"] = df["Amps"]
        validity, reason = v.validate(df)
        self.assertTrue(validity)

    def test_invalidation_maccor(self):
        path = "PredictionDiagnostics_000109_tztest.010"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator(
            schema_filename=os.path.join(
                VALIDATION_SCHEMA_DIR, "schema-maccor-2170.yaml"
            )
        )
        v.allow_unknown = True
        header = pd.read_csv(path, delimiter="\t", nrows=0)
        print(header)
        df = pd.read_csv(path, delimiter="\t", skiprows=1)
        df["State"] = df["State"].astype(str)
        df["current"] = df["Amps"]
        print(df.dtypes)
        validity, reason = v.validate(df)
        print(validity, reason)
        self.assertFalse(validity)

    def test_validate_from_paths_maccor(self):
        paths = [os.path.join(TEST_FILE_DIR, "xTESLADIAG_000019_CH70.070")]
        with ScratchDir(".") as scratch_dir:
            # Run validation on everything
            v = SimpleValidator()
            validate_record = v.validate_from_paths(
                paths,
                record_results=True,
                skip_existing=False,
                record_path=os.path.join(scratch_dir, "validation_records.json"),
            )
            df = pd.DataFrame(v.validation_records)
            df = df.transpose()
            self.assertEqual(
                df.loc["xTESLADIAG_000019_CH70.070", "method"], "simple_maccor"
            )
            self.assertEqual(df.loc["xTESLADIAG_000019_CH70.070", "validated"], True)

    def test_monotonic_cycle_index_maccor(self):
        path = "PredictionDiagnostics_000109_tztest.010"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator(
            schema_filename=os.path.join(
                VALIDATION_SCHEMA_DIR, "schema-maccor-2170.yaml"
            )
        )
        v.allow_unknown = True
        header = pd.read_csv(path, delimiter="\t", nrows=0)
        print(header)
        df = pd.read_csv(path, delimiter="\t", skiprows=1)
        df["State"] = df["State"].astype(str)
        df["current"] = df["Amps"]
        df.loc[df["Cyc#"] == 89, "Cyc#"] = 0
        validity, reason = v.validate(df)
        self.assertFalse(validity)
        self.assertEqual(reason, "cyc# needs to be monotonically increasing for processing")


@unittest.skip("Needs to be integrated with new structuring")
class ValidationEisTest(unittest.TestCase):
    # To further develop
    def setUp(self):
        pass

    def test_validation_maccor(self):
        path = "maccor_test_file_4267-66-6519.EDA0001.041"
        path = os.path.join(TEST_FILE_DIR, path)

        v = ValidatorBeep()
        v.allow_unknown = True

        df = pd.read_csv(path, delimiter="\t", skip_blank_lines=True, skiprows=10)

        self.assertTrue(v.validate_eis_dataframe(df))


@unittest.skip("Needs to be integrated with new structuring")
class SimpleValidatorTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_file_incomplete(self):
        path = "FastCharge_000025_CH8.csv"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator()
        df = pd.read_csv(path, index_col=0)

        validity, reason = v.validate(df)
        self.assertFalse(validity)
        self.assertEqual(
            reason,
            "cycle_index needs to reach at least 1 "
            "for processing, instead found:value=0.0",
        )

    def test_basic(self):
        path = "2017-05-09_test-TC-contact_CH33.csv"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator()
        df = pd.read_csv(path, index_col=0)

        validity, reason = v.validate(df)
        self.assertFalse(validity)
        self.assertEqual(
            reason,
            "Column cycle_index: integer type check failed "
            "at index 0 with value nan",
        )

        # Test bigger file, with float/numeric type
        path = "2017-08-14_8C-5per_3_47C_CH44.csv"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator()
        df = pd.read_csv(path, index_col=0)
        v.schema["cycle_index"]["schema"]["type"] = "float"
        validity, reason = v.validate(df)
        self.assertTrue(validity)
        self.assertEqual(reason, "")

        v.schema["cycle_index"]["schema"]["type"] = "numeric"
        validity, reason = v.validate(df)
        self.assertTrue(validity)
        self.assertEqual(reason, "")

        # Test good file
        path = "2017-12-04_4_65C-69per_6C_CH29.csv"
        path = os.path.join(TEST_FILE_DIR, path)

        v = SimpleValidator()
        df = pd.read_csv(path, index_col=0)
        validity, reason = v.validate(df)
        self.assertTrue(validity)
        self.assertEqual(reason, "")

        # Alter the schema on-the-fly to induce error
        v.schema["discharge_capacity"]["schema"]["max"] = 1.8
        validity, reason = v.validate(df)
        self.assertFalse(validity)
        # Cut off reasons to prevent floating format discrepancies
        self.assertEqual(
            reason[:80],
            "discharge_capacity is higher than allowed max 1.8 "
            "at index 11154: value=1.801418",
        )

        # Alter the schema on-the-fly to move on to the next errors
        v.schema["discharge_capacity"]["schema"]["max"] = 2.1
        v.schema["step_time"] = {"schema": {"min": 0.0, "type": "float"}}
        validity, reason = v.validate(df)
        self.assertFalse(validity)
        self.assertEqual(
            reason[:70],
            "step_time is lower than allowed min 0.0 " "at index 104416:value=-450.945",
        )

        # Alter schema once more to recover validation
        del v.schema["step_time"]["schema"]["min"]
        validity, reason = v.validate(df)
        self.assertTrue(validity)
        self.assertEqual(reason, "")

    def test_project_name(self):
        with ScratchDir("."):
            os.environ["BEEP_PROCESSING_DIR"] = os.getcwd()
            os.mkdir("data-share")
            os.mkdir(os.path.join("data-share", "validation"))

            v = SimpleValidator()
            paths = [
                "FastCharge_000000_CH29.csv",
                "FastCharge_000025_CH8.csv",
                "PredictionDiagnostics_000151_test.052",
                "PriorLAMVal_000101_000080_CH2.csv",
            ]
            paths = [os.path.join(TEST_FILE_DIR, path) for path in paths]
            validate_record = v.validate_from_paths(
                paths, record_results=True, skip_existing=False
            )

            self.assertEqual(
                validate_record["FastCharge_000000_CH29.csv"]["method"],
                "schema-arbin-lfp.yaml",
            )
            self.assertEqual(
                validate_record["FastCharge_000025_CH8.csv"]["method"],
                "schema-arbin-lfp.yaml",
            )
            self.assertEqual(
                validate_record["PredictionDiagnostics_000151_test.052"]["method"],
                "schema-maccor-2170.yaml",
            )
            self.assertEqual(
                validate_record["PriorLAMVal_000101_000080_CH2.csv"]["method"],
                "schema-arbin-18650-2170.yaml",
            )

    def test_validation_from_json(self):
        with ScratchDir("."):
            os.environ["BEEP_PROCESSING_DIR"] = os.getcwd()
            os.mkdir("data-share")
            os.mkdir(os.path.join("data-share", "validation"))
            paths = [
                "2017-05-09_test-TC-contact_CH33.csv",
                "2017-12-04_4_65C-69per_6C_CH29.csv",
            ]
            paths = [os.path.join(TEST_FILE_DIR, path) for path in paths]
            # Create dummy json obj
            json_obj = {
                "file_list": paths,
                "run_list": list(range(len(paths))),
            }
            json_string = json.dumps(json_obj)
            json_output = validate_file_list_from_json(json_string)
            loaded = json.loads(json_output)
        self.assertEqual(loaded["validity"][0], "invalid")
        self.assertEqual(loaded["validity"][1], "valid")

        # Workflow output
        output_file_path = Path(tempfile.gettempdir()) / "results.json"
        self.assertTrue(output_file_path.exists())

        output_json = json.loads(output_file_path.read_text())

        self.assertEqual(paths[0], output_json["filename"])
        self.assertEqual(os.path.getsize(output_json["filename"]), output_json["size"])
        self.assertEqual(0, output_json["run_id"])
        self.assertEqual("validating", output_json["action"])
        self.assertEqual("invalid", output_json["status"])

    @unittest.skipUnless(False, "toggle this test")
    def test_heavy(self):
        # Sync all S3 objects
        from beep.utils.memprof import cache_all_kitware_data

        cache_all_kitware_data()
        paths = os.listdir(os.path.join(S3_CACHE, "D3Batt_Data_publication"))
        paths = [path for path in paths if not "Metadata" in path]
        paths = [
            os.path.join(S3_CACHE, "D3Batt_Data_publication", path) for path in paths
        ]

        # Run validation on everything
        v = SimpleValidator()
        validate_record = v.validate_from_paths(
            paths, record_results=True, skip_existing=True
        )
        df = pd.DataFrame(v.validation_records)
        df = df.transpose()
        print(df)
        print(
            "{} valid, {} invalid".format(
                len([x for x in df.validated if x]),
                len([x for x in df.validated if not x]),
            )
        )
        invalid_runs = df[np.logical_not(df.validated)]
        print(invalid_runs)
        self.assertEqual(len(invalid_runs), 3)


if __name__ == "__main__":
    unittest.main()
